#include "YumIO.h"
#include "YumIOC.h"
#include <QtGui/QMessageBox>
#include <QtCore/QRegExp>
#include <QtCore/QEventLoop>
#include <QtCore/QFile>
#include "ScrollableErrorDialog.h"
#include <QtCore/qglobal.h>
#include <QtCore/QThread>
#include <QtCore/QMutex>
#include <QtCore/QTimer>
#include "Share.h"
#include <rpm/rpmlib.h>
#include <rpm/rpmcli.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "stringmacros.h"
#include "globals.h"

#undef DEBUG_COLLECT
#undef DEBUG_COLLECT_DETAIL
#undef DEBUG_TS
#undef DEBUG_TS_DETAIL
#undef DEBUG_COLLECT_GRAB_METER_ARGS
#undef DEBUG_COLLECT_GRAB_METER
#undef DEBUG_COLLECT_GRAB_METER_MULTI
#undef DEBUG_COLLECT_GRAB_METER_DETAIL
#undef DEBUG_APPLY_GRAB_METER_ARGS
#undef DEBUG_APPLY_GRAB_METER
#undef DEBUG_APPLY_GRAB_METER_MULTI
#undef DEBUG_APPLY_GRAB_METER_DETAIL
#undef DEBUG_EXPECTED_EXCEPTIONS
#undef DEBUG_ENVRA
#undef DEBUG_DISABLE_CAPTURERS

#if PY_VERSION_HEX < 0x02050000
typedef int  Py_ssize_t;
#endif

//
// Values passed to output.YumCliRPMCallback by rpm in yum >= 3.2.3
//
#define TS_UPDATE       10
#define TS_INSTALL      20
#define TS_TRUEINSTALL  30
#define TS_ERASE        40
#define TS_OBSOLETED    50
#define TS_OBSOLETING   60
#define TS_AVAILABLE    70
#define TS_UPDATED      90

#define IS_INSTALL(action) \
   ((action) == TS_INSTALL || (action) == TS_TRUEINSTALL)
#define IS_UPDATE(action) \
   ((action) == TS_UPDATE || (action) == TS_OBSOLETING)
#define IS_REMOVE(action) \
   ((action) == TS_ERASE)
#define IS_COMPLETING_UPDATE(action) \
   ((action) == TS_UPDATED || (action) == TS_OBSOLETED)

//
// Known yum strings for collect's progress indications
//
#define COLL_GRAB_METADATA  "MD Read"

//
// Known yum strings for apply's depsolve
//
// I hate having to know these strings, but there is no callback for a
// package added to the transaction by depSolve.  (There's one for the
// Requires: string that caused it, but that's not good enough, because
// the Requires: might not mention the package name but just something
// else Provided by the package.)
//
// The actual strings look like:
//    TSINFO: Updating %s to resolve dep.
//    TSINFO: Updating %s to resolve conflict.
//    TSINFO: Obsoleting %s with %s to resolve dep.
//    TSINFO: Marking %s as update for %s
//    TSINFO: Marking %s as install for %s
//    TSINFO: %s package requiring %s marked as erase
//
#define APPLY_DEPSOLVE_PREFIX           "TSINFO: "
#define APPLY_DEPSOLVE_UPDATING         "Updating "
#define APPLY_DEPSOLVE_OBSOLETING       "Obsoleting "
#define APPLY_DEPSOLVE_OBSOLETING_WITH  " with "
#define APPLY_DEPSOLVE_MARKING          "Marking "
#define APPLY_DEPSOLVE_MARKING_UPDATE   " update "
#define APPLY_DEPSOLVE_ERASE            " marked as erase"

//
// Known yum strings for available packages
//
#define YUM_DECODE_PRCO_FLAGS(f)  (strcmp((f), "GT") == 0 ? ">"  : \
                                   strcmp((f), "GE") == 0 ? ">=" : \
                                   strcmp((f), "LT") == 0 ? "<"  : \
                                   strcmp((f), "LE") == 0 ? "<=" : \
                                   strcmp((f), "EQ") == 0 ? "="  : \
                                                            (f))

// When using package sizes to estimate the amount of time for a package,
// very small packages still will have overhead time.  So, use a lower bound
// of PKG_MIN as the "size" of those packages when figuring the total
// progress.  (Continue to use the actual size for item progress, though.)
#define PKG_MIN  10240ULL

#define RA_IMAGE(act)                         \
   (act == RA_Grab_Metalink ? "Grab_Meta"   : \
    act == RA_Grab_RepoMD   ? "Grab_RepoMD" : \
    act == RA_Grab_Group    ? "Grab_Group"  : \
    act == RA_Grab          ? "Grab"        : \
    act == RA_Convert       ? "Convert"     : \
    act == RA_Read          ? "Read"        : \
                              "?")

//
// Progress values for collect
//
#define PG_C_INITIALIZE     100
#define PG_C_REPOS_START    900
#define PG_C_REPOS        14000
#define PG_C_UPDATES       4000
#define PG_C_INSTALL       4000
#define PG_C_GROUPS        1000

#define PG_C_TOTAL_TEXT        "Initializing: Total Progress"
#define PG_C_INITIALIZE_TEXT   "Preparing"
#define PG_C_REPOS_START_TEXT  "Reading repositories"
//      PG_C_REPOS_TEXT        "Downloading/Reading repository metadata for ..."
#define PG_C_UPDATES_TEXT      "Collecting updatable packages"
#define PG_C_INSTALL_TEXT      "Collecting installed and available packages"
#define PG_C_GROUPS_TEXT       "Collecting groups"

#define PG_C_TOTAL  (  PG_C_INITIALIZE  \
                     + PG_C_REPOS_START \
                     + PG_C_REPOS       \
                     + PG_C_UPDATES     \
                     + PG_C_INSTALL     \
                     + PG_C_GROUPS)

//
// Progress values for collectFileLists
//
#define PG_F_REPOS_START   1000
#define PG_F_REPOS        14000

#define PG_F_TOTAL_TEXT        "Loading file lists: Total Progress"
#define PG_F_REPOS_START_TEXT  "Reading repositories"
//      PG_F_REPOS_TEXT        "Downloading/Reading repository metadata for ..."

#define PG_F_TOTAL  (  PG_F_REPOS_START \
                     + PG_F_REPOS)

//
// Progress values for collectOtherData
//
#define PG_O_REPOS_START   1000
#define PG_O_REPOS        14000

#define PG_O_TOTAL_TEXT        "Loading miscellaneous data: Total Progress"
#define PG_O_REPOS_START_TEXT  "Reading repositories"
//      PG_O_REPOS_TEXT        "Downloading/Reading repository metadata for ..."

#define PG_O_TOTAL  (  PG_O_REPOS_START \
                     + PG_O_REPOS)

//
// Progress values for apply's Check Dependencies (i.e. before confirmation)
//
#define PG_D_INITIALIZE    100
#define PG_D_DEPSOLVE     9900

#define PG_D_TOTAL_TEXT        "Checking Dependencies: Total Progress"
#define PG_D_INITIALIZE_TEXT   "Preparing"
//      PG_D_DEPSOLVE_TEXT     "Downloading ... header"
//                             "Resolve Dependencies"


#define PG_D_TOTAL        (  PG_D_INITIALIZE  \
                           + PG_D_DEPSOLVE)

//
// Progress values for apply's Run Transaction (i.e. after confirmation)
//
#define PG_R_INITIALIZE           100
#define PG_R_GRAB_RPMS_NOMINAL  24000 // used to define ts.run.pgRGrabRpms
#define PG_R_DELTA_REBUILD       2000
#define PG_R_TRANS_TEST_INIT     1000
#define PG_R_TRANS_TEST          1000
#define PG_R_RUN_TRANS_INIT      1000
#define PG_R_RUN_TRANS           8000

#define PG_R_TOTAL_TEXT            "Running Transaction: Total Progress"
#define PG_R_INITIALIZE_TEXT       "Preparing"
//      PG_R_GRAB_RPMS             "Downloading ... header"
#define PG_R_DELTA_REBUILD_TEXT    "Finishing RPM rebuild from deltas"
#define PG_R_TRANS_TEST_INIT_TEXT  "Preparing to Test Transaction"
//      PG_R_TRANS_TEST            "Transaction Test"
#define PG_R_RUN_TRANS_INIT_TEXT   "Preparing to Run Transaction"
//      PG_R_RUN_TRANS             "Transaction Test"
//                                 "Installing: ..."
//                                 "Updating: ..."
//                                 "Removing: ..."
//                                 "Completing update for: ..."

#define PG_R_TOTAL        (  PG_R_INITIALIZE                                  \
                           + ts.run.pgRGrabRpms                               \
                           + (usesPresto() ? PG_R_DELTA_REBUILD : 0)          \
                           + (hasTransTest()                                  \
                              ? (PG_R_TRANS_TEST_INIT + PG_R_TRANS_TEST) : 0) \
                           + PG_R_RUN_TRANS_INIT                              \
                           + PG_R_RUN_TRANS)

// The number of acts given the number of grabs
#define PG_D_GRAB(grabs)  ((grabs) * 10)

// Estimate that a dependency resolution is 10% of the total time for
// grabbing the headers
#define PG_D_DEPSOLVE_ESTIMATE(pkgs)  ((pkgs) * 10 / 10)

// The number of acts for each download based on the size of the package
#define PG_R_GRAB(size)   (qMax((size), PKG_MIN))

// The number of acts for each operation based on the size of the package
//
// The size of an installed package is reported in uncompressed bytes.
// But the size of an uninstalled package is reported as its file size.
// Convert uninstalled package sizes by multiplying by 3 (a heuristic value).
//
// The COMPLETE_UPDATE is the erase part of the update.  It usually is pretty 
// fast because most of the time it doesn't have to delete much (or anything).
// So divide its size by 5 (a heuristic value).
#define PG_R_INSTALL(size)          (qMax((size) * 3, PKG_MIN))
#define PG_R_REMOVE(size)           (qMax((size),     PKG_MIN))
#define PG_R_UPDATE(size)           (qMax((size) * 3, PKG_MIN))
#define PG_R_COMPLETE_UPDATE(size)  (qMax((size) / 5, PKG_MIN))

// Estimate that a transaction test is 5% of the total time for the main
// part of the transaction
#define PG_R_TEST_TRANS_ESTIMATE(acts)  ((acts) / 20)

class YumIO::EntryKey {
public:
   EntryKey(bool     isGroup,
            QString  name,
            QString  biarch)
      : isGroup(isGroup), 
        name(name),
        biarch(biarch)
   {}
   EntryKey(Package::Kind  kind,
            QString        name,
            QString        biarch)
      : isGroup(kind == Package::GROUP), 
        name(name),
        biarch(biarch)
   {}
   inline bool operator==(EntryKey&  that) const
   {
      return (this->isGroup == that.isGroup
              && this->name == that.name
              && this->biarch == that.biarch);
   }
   bool     isGroup;
   QString  name;
   QString  biarch;
};

inline uint qHash(const YumIO::EntryKey&  key)
{
   extern uint qHash(const QString &key);
   return (qHash(key.name) << 1) + (qHash(key.biarch) << 2) + key.isGroup;
}

YumIO::YumIO()
   : clearing(false),
     logStartOp(-1),
     collectedFileLists(false),
     collectedOtherData(false),
     dialogs(NULL),
     capturers(NULL),
     rpm_dict(NULL),
     classYum(NULL),
     yumIO(NULL),
     reposDict(NULL),
     locked(false),
     yumVerKnown(false),
     usesSqliteKnown(false),
     usesSqliteValue(false),
     usesPrestoKnown(false),
     usesPrestoValue(false),
     hasMultiProgressKnown(false),
     hasMultiProgressValue(false),
     totalProgressBar(NULL),
     itemProgressBar(NULL),
     pg(NULL)
{
   wr.stderrKind  = Log::ERROR;
   wr.stderrValue = 0;
}

YumIO::~YumIO()
{
   Python::PyGILStateWrapper  gsw;

   if (initialized) {
      // Since the log window is toast now, and we're destroying the yumIO
      // now, which will render the YumIOStdout and YumIOStderr python
      // objects useless, let's change the python stdout & stderr to their
      // defaults.  That way, if any errors happen, they'll be seen and won't
      // cause either cascading hangs or SIGSEGV's.
      py("sys.stdout = sys.__stdout__\n"
         "sys.stderr = sys.__stderr__\n");
   }

   // To be proper, we'd call:
   //    Py_DECREF(yumIO);
   //    Py_DECREF(reposDict);
   // But this happens on exit, so don't waste the time.

   if (locked) {
      pyExceptStart();
      pyExceptExpect(xcpt.lockError);
      PyObject*  unlockStatus = pyMethodCall(yumIO, "unlock", "()");
      if (pyExceptMatches(unlockStatus, xcpt.lockError)) {
         pythonException(tr("Unable to release yum lock"));
      }
      Py_XDECREF(unlockStatus);
      pyExceptClear();
   }
}

void YumIO::setup(QWidget*       parent,
                  HProgressBar*  totalProgressBar,
                  HProgressBar*  itemProgressBar)
{
   dialogs = new YumIODialogs(parent);
   this->parent           = parent;
   this->totalProgressBar = totalProgressBar;
   this->itemProgressBar  = itemProgressBar;
}

bool YumIO::initialize()
{
   Python::initialize();
   if (yumIO == NULL) {
      Python::PyGILStateWrapper  gsw;

      YumIOC::initialize(this);

      // This is necessary for newer versions of yum (e.g. 2.4.2) when
      // calling getOptionsConfig in python.  And it's harmless for older
      // versions of yum (e.g. 2.2.2).
      PyObject*  yumArgv = Py_BuildValue("[s,s]", program, "info");
      pyAssign("sys", "argv", yumArgv);
      Py_DECREF(yumArgv);

      PyObject*  rpm_mod  = PyImport_ImportModule("rpm");
      if (rpm_mod == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: cannot find python rpm module"));
         return false;
      }

      rpm_dict = PyModule_GetDict(rpm_mod);
      if (rpm_dict == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: cannot find python rpm dict"));
         return false;
      }

      setEnablePythonSIGINTHandler(false);

      FILE*  yumIOFile
         = fopen(Share::absolute("python/YumIO.py").toStdString().c_str(), "r");
      if (yumIOFile == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: unable to open YumIO.py"));
         return false;
      }
      int  status = PyRun_SimpleFile(yumIOFile, "YumIO.py");
      if (status == -1) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: unable to execute YumIO.py"));
         return false;
      }
      fclose(yumIOFile);

      classYum = PyDict_GetItemString(main_dict, "YumIO");
      if (classYum == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: cannot find python YumIO"));
         return false;
      }
      pyExceptStart();
      pyExceptExpect(PyExc_SystemExit);
      PyObject*  args = Py_BuildValue("()");
      yumIO = PyObject_Call(classYum, args, NULL);
      Py_XDECREF(args);
      if (pyExceptMatches(yumIO, PyExc_SystemExit)) {
         pythonException(tr("Unable to initialize"));
         return false;
      } else if (yumIO == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: cannot create python yumIO"));
         return false;
      }

      xcpt.bdbQuit                 = pyLookup("bdb.BdbQuit");
      xcpt.yumBaseError            = pyLookup("yum.Errors.YumBaseError");
      xcpt.lockError               = pyLookup("yum.Errors.LockError");
      xcpt.groupsError             = pyLookup("yum.Errors.GroupsError");
      xcpt.repoError               = pyLookup("yum.Errors.RepoError");
      xcpt.sqlite3OperationalError = pyLookup("sqlite3.OperationalError",
                                              /* tolerant => */ true);
      xcpt.noMoreMirrorsRepoError
         = pyLookup("yum.Errors.NoMoreMirrorsRepoError",
                    /* tolerant => */ true);
   }

   return true;
}

void  YumIO::blockerRelease ()
{
   blocker.release();
}

void  YumIO::execWorkThread (QEventLoop&  eventLoop,
                             QThread&     thread)
{
   blocker.prepare();

   PyThreadStateWrapper  tsw;

   thread.start(QThread::HighPriority);

   sigset_t  blocked;
   sigset_t  original;
   sigemptyset(&blocked);
   sigaddset(&blocked, SIGCHLD);
   sigprocmask(SIG_BLOCK, &blocked, &original);

   // Release the blocker once the eventLoop starts
   QTimer::singleShot(0, this, SLOT(blockerRelease()));

   eventLoop.exec();
   thread.wait();
   eventLoop.processEvents(QEventLoop::ExcludeUserInputEvents);

   // Undo the obnoxious SIGINT handler that either yum or rpm installed again
   {
      Python::PyGILStateWrapper  gsw;
      setEnablePythonSIGINTHandler(false);
   }

   sigprocmask(SIG_SETMASK, &original, NULL);
}

class CollectThread : public QThread 
{
public:
   CollectThread(YumIO*               yumIO,
                 QEventLoop&          eventLoop,
                 const YumReposUsed&  reposUsed)
      : yumIO(yumIO),
        eventLoop(eventLoop),
        reposUsed(reposUsed)
   {}
protected:
   YumIO*               yumIO;
   QEventLoop&          eventLoop;
   const YumReposUsed&  reposUsed;
public:
   bool                 result;
protected:
   void run()
   {
      result = yumIO->collectReal(reposUsed);
      eventLoop.exit();
   }
};

bool YumIO::collect (const YumReposUsed&  reposUsed)
{
   QEventLoop     eventLoop;
   CollectThread  thread(this, eventLoop, reposUsed);
   execWorkThread(eventLoop, thread);
   return thread.result;
}

bool YumIO::collectReal (const YumReposUsed&  reposUsed)
{
   blocker.block();

   Progress::User    pgUser(pg, totalProgressBar, itemProgressBar);
   YumIODialogsUser  yumIODialogsUser(dialogs, parent);

   log(Log::ACTION, 0, tr("Initialization"));

   Python::PyGILStateWrapper  gsw;

   collectedFileLists = false;
   collectedOtherData = false;

   pg->startTotal(tr(PG_C_TOTAL_TEXT), PG_C_TOTAL);

   pg->startItem(tr(PG_C_INITIALIZE_TEXT), PG_C_INITIALIZE);
   if (!locked) {
      pyExceptStart();
      pyExceptExpect(xcpt.lockError);
      PyObject*  lockStatus = pyMethodCall(yumIO, "lock", "()");
      if (pyExceptMatches(lockStatus, xcpt.lockError)) {
         pythonException(tr("Unable to obtain yum lock"));
         pyExceptClear();
         pg->reset();
         return false;
      }
      Py_XDECREF(lockStatus);
      pyExceptClear();

      locked = true;
   }

   PyObject*  repos     = PyObject_GetAttrString(yumIO, "repos");
              reposDict = PyObject_GetAttrString(repos, "repos");
   if (reposDict == NULL) {
      error(tr("Internal error: yumIO.repos.repos not evaluable"));
      return false;
   }
   if (!PyDict_Check(reposDict)) {
      error(tr("Internal error: yumIO.repos.repos not a dictionary"));
      return false;
   }
   
   useRepos(repos, reposUsed);

   Py_DECREF(repos);

   repoProgress.repos = collectRepos(reposDict);
   repoProgress.kind  = tr("contents");
   repoProgress.act   = 0;

   repoProgress.parallel.begun   = false;
   repoProgress.parallel.setup   = false;
   repoProgress.parallel.meterMapping.clear();
   repoProgress.parallel.threads = NULL;
   repoProgress.parallel.retryMapping.clear();

   if (PyDict_Size(reposDict) == 0) {
      error(tr("No repositories are configured.\n"
               "Under the Repositories menu, Edit Configuration can be used"
               " to configure repositories."));
      return false;
   } else if (repoProgress.repos.count() == 0) {
      error(tr("No repositories are enabled.\n"
               "Under the Repositories menu, Edit Configuration can be used"
               " to configure repositories."));
      return false;
   }

   pg->startItem(tr(PG_C_REPOS_START_TEXT), PG_C_REPOS_START);

   pg->setupItemActs(PG_C_REPOS, repoProgress.repos.count() * RA_Count);
   pyExceptStart();
   pyExceptExpect(xcpt.repoError);
   pyExceptExpect(xcpt.bdbQuit);
   PyObject*  reposStatus = pyMethodCall(yumIO, "setupRepos", "()");
   if (pyExceptMatches(reposStatus, xcpt.repoError)
       || pyExceptMatches(reposStatus, xcpt.bdbQuit)) {
      pythonExceptionV(tr("Unable to initialize"));
      pyExceptClear();
      pg->reset();
      return false;
   }
   Py_XDECREF(reposStatus);
   pyExceptClear();
   pg->endItemActs();

   basearch = getYumVar("basearch");

   pg->startItem(tr(PG_C_UPDATES_TEXT), PG_C_UPDATES);
   pyExceptStart();
   commExceptSetup();
   PyObject*  setupTsStatus = pyMethodCall(yumIO, "setupTs", "()");
   if (commExceptRaised(setupTsStatus, "setup transaction")) {
      pyExceptClear();
      pg->reset();
      return false;
   }
   Py_XDECREF(setupTsStatus);
   pyExceptClear();
   PyObject*  updates;
   {
#ifndef DEBUG_DISABLE_CAPTURERS
      YumIOCapturersUser  capturersUser(capturers, this);
#endif
      pyExceptStart();
      commExceptSetup();
      updates = pyMethodCall(yumIO, "collectUpdates", "()");
      if (commExceptRaised(updates, "collect updatable packages")) {
         pyExceptClear();
         pg->reset();
         return false;
      } else if (updates == NULL) {
         error(tr("Internal error: execution of yumIO.collectUpdates failed"));
         pyExceptClear();
         pg->reset();
         return false;
      }
      pyExceptClear();
   }

   if (!collectSet(hash, updates, Package::OUT_OF_DATE)) {
      return false;
   }

   Py_DECREF(updates);

   pg->startItem(tr(PG_C_INSTALL_TEXT), PG_C_INSTALL);
   PyObject*  info;
   {
#ifndef DEBUG_DISABLE_CAPTURERS
      YumIOCapturersUser  capturersUser(capturers, this);
#endif
      pyExceptStart();
      commExceptSetup();
      info = pyMethodCall(yumIO, "collectAll", "()");
      if (commExceptRaised(info, "collect installed and available packages")) {
         pyExceptClear();
         pg->reset();
         return false;
      } else if (info == NULL) {
         error(tr("Internal error: execution of yumIO.collectAll failed"));
         pyExceptClear();
         pg->reset();
         return false;
      }
      pyExceptClear();
   }

   if (   !collectField(hash, info, "installed", Package::INSTALLED)
       || !collectField(hash, info, "available", Package::NOT_INSTALLED)) {
      return false;
   }

   Py_DECREF(info);

   pg->startItem(tr(PG_C_GROUPS_TEXT), PG_C_GROUPS);
   pyExceptStart();
   pyExceptExpect(xcpt.groupsError);
   commExceptSetup();
   PyObject*  setupTsGroupResult = pyMethodCall(yumIO, "setupTsGroup", "()");
   if (pyExceptMatches(setupTsGroupResult, xcpt.groupsError)) {
      QString  value = pyStringDec(pyExceptValuef());
      if (value.contains("No Groups Available")) {
         // DO NOTHING: It's fairly normal.  (Although why it's lumped with 
         //             actually serious errors as the single 
         //             yum.Errors.GroupsErrors exception I have no idea!)
      } else {
         error(tr("Internal error:"
                  " execution of yumIO.setupTsGroup failed"));
         pyExceptClear();
         pg->reset();
         return false;
      }
   } else if (commExceptRaised(setupTsGroupResult, "collect groups")) {
      pyExceptClear();
      pg->reset();
      return false;
   }
   Py_XDECREF(setupTsGroupResult);
   pyExceptClear();
   PyObject*  groupInfo = pyMethodCall(yumIO, "collectGroupInfo", "()");
   if (groupInfo == NULL) {
      error(tr("Internal error: execution of yumIO.collectGroupInfo failed"));
      return false;
   }

   if (!collectGroups(hash, groupInfo)) {
      return false;
   }

   PyObject*  collectDoneStatus = pyMethodCall(yumIO, "collectDone", "()");
   Py_DECREF(collectDoneStatus);

   Py_DECREF(groupInfo);

   EntryHash::const_iterator  it = hash.begin();
   for (; it != hash.end(); ++it) {
      list.append(*it);
   }

   return true;
}

void  YumIO::useRepos(PyObject*            repos,
                      const YumReposUsed&  reposUsed)
{
   YumReposUsed::const_iterator  it = reposUsed.begin();
   for (; it != reposUsed.end(); ++it) {
      PyObject*  repo = pyDictGetItem(reposDict, it.key());
      if (repo != NULL) {
         char*  id = strdup(it.key().toStdString().c_str());
         pyExceptStart();
         pyExceptExpect(xcpt.repoError);
         PyObject*  status;
         if (*it) {
            status = pyMethodCall(repos, "enableRepo", "(s)", id);
         } else {
            status = pyMethodCall(repos, "disableRepo", "(s)", id);
         }
         if (pyExceptMatches(status, xcpt.repoError)) {
            warning(tr("unable to setup repository %1:\n%2")
                       .arg(it.key()).arg(pyStringDec(pyExceptValuef())));
         }
         Py_XDECREF(status);
         pyExceptClear();
         free(id);
      }
   }
}

QStringList  YumIO::collectRepos (PyObject*  dict)
{
   QStringList  result;

   Py_ssize_t  pos = 0;
   for (;;) {
      PyObject*  key;
      PyObject*  value;
      if (!PyDict_Next(dict, &pos, &key, &value)) break;

      PyObject*  enabled = PyObject_GetAttrString(value, "enabled");
      if (enabled != NULL && PyInt_AsLong(enabled) != 0) {
         result.append(pyString(key));
      }
      Py_XDECREF(enabled);
   }

   return result;
}

bool  YumIO::collectField (EntryHash&      hash,
                           PyObject*       info,
                           const char*     field,
                           Package::State  state)
{
   char*  copy = new char[strlen(field)+1];
   strcpy(copy, field);
   bool  result = collectField(hash, info, copy, state);
   delete [] copy;
   return result;
}

bool  YumIO::collectField (EntryHash&      hash,
                           PyObject*       info,
                           char*           field,
                           Package::State  state)
{
   PyObject*  set = PyObject_GetAttrString(info, field);
   if (set == NULL) {
      error(tr("Internal error: yumIO.collectAll.%1 not found").arg(field));
      return false;
   }

   bool  result = collectSet(hash, set, state);
   Py_DECREF(set);
   return result;
}

bool  YumIO::collectSet (EntryHash&      hash,
                         PyObject*       set,
                         Package::State  state)
{
   PyObject*  it = PyObject_GetIter(set);
   for (;;) {
      PyObject*  pkg = PyIter_Next(it);
      if (pkg == NULL) break;

      QString  name(pyStringDec(PyObject_GetAttrString(pkg, "name")));
      QString  arch(pyStringDec(PyObject_GetAttrString(pkg, "arch")));
      QString  biarch(getBiArch(arch));

      biarches.insert(biarch);

      EntryKey  key(false, name, biarch);
      if (!hash.contains(key)) {
         Package::Entry*&  entry = hash[key];
         bool  status = collectPkg(entry, state, pkg);
         if (!status) return false;
         if (namesBest.contains(name)) {
            Package::Entry*&  best = namesBest[name];
            PyObject*  result = pyMethodCall(yumIO, "isBetterThan", "(O,O)", 
                                             best->obj, entry->obj);
            if (result == Py_True) {
               entry->duplicateName = true;
            } else {
               best->duplicateName = true;
               best = entry;
            }
            Py_XDECREF(result);
         } else {
            namesBest[name] = entry;
         }
      } else {
         Package::Entry*  entry = hash[key];
         if (entry->state == Package::OUT_OF_DATE
             && state == Package::INSTALLED) {
            // For an out-of-date package, in addition to the version
            // available for installation, keep track of the installed
            // version.  The Describe dialog displays both.
            while (entry->next != NULL) {
               entry = entry->next;
            }
            collectPkg(entry->next, state, pkg);
         }
      }

      Py_DECREF(pkg);
   }

   Py_DECREF(it);

   return true;
}

bool  YumIO::collectPkg (Package::Entry*&  entry,
                         Package::State    state,
                         PyObject*         pkg)
{
   entry = new Package::Entry;
   
   Py_INCREF(pkg);

   entry->action        = Package::NONE;
   entry->state         = state;
   entry->kind          = getPkgKind(pkg);
   entry->name          = pyStringDec(PyObject_GetAttrString(pkg, "name"));
   entry->verrel        = pyStringDec(pyMethodCall(pkg, "printVer", "()"));
   entry->repo          = simpleString(pkg, "repoid");
   entry->arch          = pyStringDec(PyObject_GetAttrString(pkg, "arch"));
   entry->biarch        = getBiArch(entry->arch);
   entry->size          = sizeOf(pkg);
   entry->sizeText      = bytesFormat(entry->size);
   entry->summary       = simpleString(pkg, "summary");
   entry->buildTime     = buildTimeOf(pkg);
   entry->buildTimeText = timeFormat(entry->buildTime);
   entry->obj           = pkg;

   return true;
}

bool  YumIO::collectGroups (EntryHash&  hash,
                            PyObject*   groupInfo)
{
   PyObject*  list;

   if (hasDistribGroups()) {
      list = groupInfo;
   } else {
      list = PyObject_GetAttrString(groupInfo, "visible_groups");
      if (list == NULL) {
         error(tr("Internal error: groupInfo.visible_groups not evaluable"));
         return false;
      }
   }

   PyObject*  it = PyObject_GetIter(list);
   for (;;) {
      PyObject*  obj = PyIter_Next(it);
      if (obj == NULL) break;

      PyObject*  nameObj;
      PyObject*  groupObj;
      if (hasDistribGroups()) {
         PyObject*   user_visible = PyObject_GetAttrString(obj, "user_visible");
         if (user_visible == NULL || !PyObject_IsTrue(user_visible)) continue;
         nameObj = PyObject_GetAttrString(obj, "name");
         if (nameObj == NULL) {
            error(tr("Internal error: group without name"));
            return false;
         }
         groupObj = obj;
      } else {
         nameObj  = obj;
         groupObj = groupInfo;
      }

      QString   name = pyString(nameObj);
      EntryKey  key(true, name, QString());

      if (!hash.contains(key)) {
         bool  status = collectGroup(hash[key], groupObj, name);
         if (!status) return false;
      }

      if (hasDistribGroups()) {
         Py_DECREF(nameObj);
      }
      
      Py_DECREF(obj);
   }

   Py_DECREF(it);

   if (!hasDistribGroups()) {
      Py_DECREF(list);
   }

   return true;
}

bool  YumIO::collectGroup (Package::Entry*&  entry,
                           PyObject*         groupObj,
                           QString           name)
{
   bool  installed;

   if (hasDistribGroups()) {
      // groupObj is a yum.comps.Group
      PyObject*  installedObj = PyObject_GetAttrString(groupObj, "installed");
      if (installedObj == NULL) {
         error(tr("Internal error: cannot find group's installed flag"));
         return false;
      }
      installed = PyObject_IsTrue(installedObj);
      Py_DECREF(installedObj);
   } else {
      // groupObj is a groupInfo
      PyObject*  groupInstalled = PyObject_GetAttrString(groupObj,
                                                         "group_installed");
      char*      nameCopy       = strdup(name.toStdString().c_str());
      PyObject*  installedObj   = PyDict_GetItemString(groupInstalled, nameCopy);
      free(nameCopy);
      if (installedObj == NULL) {
         error(tr("Internal error:"
                  " cannot find groupInfo.group_installed['%s']").arg(name));
         return false;
      }
      Py_DECREF(groupInstalled);
      installed = PyObject_IsTrue(installedObj);
   }

   Py_INCREF(groupObj);

   entry = new Package::Entry;
   
   entry->action = Package::NONE;
   entry->state  = (installed ? Package::INSTALLED 
                              : Package::NOT_INSTALLED);
   entry->kind   = Package::GROUP;
   entry->name   = name;
   entry->obj    = groupObj;

   if (entry->state == Package::INSTALLED && !upToDateGroup(*entry)) {
      entry->state = Package::OUT_OF_DATE;
   }

   return true;
}

void  YumIO::clear()
{
   {
      Python::PyGILStateWrapper  gsw;

      PyObject*  flushStatus;
      flushStatus = pyMethodCall(yumIO, "flush", "()");
      Py_XDECREF(flushStatus);
   }

   clearEntries();

   {
      Python::PyGILStateWrapper  gsw;

      Py_DECREF(yumIO);
      PyObject*  args = Py_BuildValue("()");
      yumIO = PyObject_Call(classYum, args, NULL);
      Py_DECREF(args);
   }
}

class ClearEntriesThread : public QThread 
{
public:
   ClearEntriesThread(YumIO*       yumIO,
                      QEventLoop&  eventLoop)
      : yumIO(yumIO),
        eventLoop(eventLoop)
   {}
protected:
   YumIO*       yumIO;
   QEventLoop&  eventLoop;
protected:
   void run()
   {
      yumIO->clearEntriesReal();
      eventLoop.exit();
   }
};

void YumIO::clearEntries () 
{
   clearing = true;

   QEventLoop          eventLoop;
   ClearEntriesThread  thread(this, eventLoop);
   execWorkThread(eventLoop, thread);
}

void YumIO::clearEntriesReal ()
{
   blocker.block();

   Python::PyGILStateWrapper  gsw;

   {
      Package::EntryList::const_iterator  it = list.begin();
      for (; it != list.end(); ++it) {
         Package::Entry*  entry = *it;

         for (; entry != NULL; entry = entry->next) {
            Py_XDECREF(entry->obj);
            entry->obj = NULL;
         }
      }
   }

   qDeleteAll(list.begin(), list.end());
   list.clear();
   hash.clear();
   namesBest.clear();

   biarches.clear();

   clearing = false;
}

void YumIO::clearEntriesBeforeQuit () 
{
   // This leaks all the Package::Entry python objects, but it's no problem
   // since we're quitting anyway.

   clearing = true;

   Package::EntryList::const_iterator  it = list.begin();
   for (; it != list.end(); ++it) {
      Package::Entry*  entry = *it;

      for (; entry != NULL; entry = entry->next) {
         entry->obj = NULL;
      }
   }

   qDeleteAll(list.begin(), list.end());
   list.clear();
   hash.clear();
   namesBest.clear();

   biarches.clear();
}

QString  YumIO::getState (const Package::Entry&  entry)
{
   switch (entry.state) {
   case Package::OUT_OF_DATE:   return tr("Available for update");
   case Package::NOT_INSTALLED: return tr("Available for installation");
   case Package::INSTALLED:     return tr("Installed");
   default:                     return QString();
   }
}

QString  YumIO::getKind (const Package::Entry&  entry)
{
   switch (entry.kind) {
   case Package::SUITE:   return tr("Suite of packages");
   case Package::GROUP:   return tr("Group of packages");
   case Package::PACKAGE: return tr("Ordinary package");
   default:               return QString();
   }
}

QString  YumIO::getDescription (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "description");
}

QString  YumIO::getGroup (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "group");
}

QString  YumIO::getVendor (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "vendor");
}

QString  YumIO::getBuildHost (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "buildhost");
}

QString  YumIO::getInstallTime (const Package::Entry&  entry)
{
   return getRpmTime(entry.obj, "INSTALLTIME");
}

QString  YumIO::getBinaryRpm (const Package::Entry&  entry)
{
   // In newer yum versions, the relativepath field doesn't even exist.
   {
      Python::PyGILStateWrapper  gsw;
      if (!PyObject_HasAttrString(entry.obj, "relativepath")) return QString();
   }

   return simpleString(entry.obj, "relativepath");
}

QString  YumIO::getSourceRpm (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "sourcerpm");
}

QString  YumIO::getLicense (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "license");
}

QString  YumIO::getPackager (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "packager");
}

QString  YumIO::getUrl (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "url");
}

QString  YumIO::getRelocations (const Package::Entry&  entry)
{
   return getRpmList(entry.obj, "PREFIXES");
}

QString  YumIO::getProvides (const Package::Entry&  entry)
{
   return getPrco(entry.obj, "PROVIDE", "provides");
}

QString  YumIO::getRequires (const Package::Entry&  entry)
{
   return getPrco(entry.obj, "REQUIRE", "requires");
}

QString  YumIO::getConflicts (const Package::Entry&  entry)
{
   return getPrco(entry.obj, "CONFLICT", "conflicts");
}

QString  YumIO::getObsoletes (const Package::Entry&  entry)
{
   return getPrco(entry.obj, "OBSOLETE", "obsoletes");
}

QString  YumIO::getFiles (const Package::Entry&  entry)
{
   Python::PyGILStateWrapper  gsw;

   QStringList  values;

   switch (entry.kind) {
   case Package::SUITE:
   case Package::PACKAGE:
      if (PyObject_HasAttrString(entry.obj, "hdr")) {
         // Must be a YumInstalledPackage.  Get info from the hdr

         PyObject*  baseNamesList  = getRpmHdrValue(entry.obj, "BASENAMES");
         if (baseNamesList == NULL || baseNamesList == Py_None) {
            return QString();
         }
         PyObject*  dirNamesList   = getRpmHdrValue(entry.obj, "DIRNAMES");
         PyObject*  dirIndexesList = getRpmHdrValue(entry.obj, "DIRINDEXES");

         PyObject*  baseNamesIt  = PyObject_GetIter(baseNamesList);
         PyObject*  dirIndexesIt = (PyList_Check(dirIndexesList)
                                    ? PyObject_GetIter(dirIndexesList)
                                    : NULL);
         for (;;) {
            PyObject*  baseNameItem = PyIter_Next(baseNamesIt);
            if (baseNameItem == NULL) break;
            PyObject*  dirIndexItem = (dirIndexesIt != NULL
                                       ? PyIter_Next(dirIndexesIt)
                                       : dirIndexesList);

            QString        baseName = pyString(baseNameItem);
            unsigned long  dirIndex = PyInt_AsLong(dirIndexItem);

            PyObject*  dirNameItem = (dirNamesList != NULL
                                      ? PyList_GetItem(dirNamesList, dirIndex)
                                      : NULL);
            QString    dirName     = pyString(dirNameItem);

            values.append(QString(dirName).append(baseName));

            Py_DECREF(dirIndexItem);
            Py_DECREF(baseNameItem);
         }      

         Py_XDECREF(dirIndexesIt);
         Py_DECREF(baseNamesIt);

         Py_DECREF(dirIndexesList);
         Py_DECREF(dirNamesList);
         Py_DECREF(baseNamesList);
      } else if (PyObject_HasAttrString(entry.obj, "_loadFiles")
                 || PyObject_HasAttrString(entry.obj, "files")) {
         // Must be a YumAvailablePackageSqlite (_loadFiles) or
         // YumAvailablePackage (files).
         //
         // NOTE: If this was a YumAvailablePackageSqlite, it was crucial that
         //       we not reference "files" above, because that would cause it
         //       to be collected early.
         //
         // First make sure we've collected the File Lists, and then get the
         // files from returnFileEntries()
         bool  success;

         gsw.release();
         success = collectFileLists();
         gsw.ensure();

         if (success) {
            PyObject*  list = pyMethodCall(entry.obj, "returnFileEntries", 
                                           "(s)", "file");
            if (list == NULL) {
               error(tr("Internal error: cannot find package's files"));
               return QString();
            }

            PyObject*  it = PyObject_GetIter(list);
            for (;;) {
               PyObject*  item = PyIter_Next(it);
               if (item == NULL) break;

               values.append(pyString(item));

               Py_DECREF(item);
            }

            Py_DECREF(it);
            Py_DECREF(list);
         }
      }
      break;
   case Package::GROUP:
      // DO NOTHING
      break;
   }

   return makeSortedUniq(values);
}

QString  YumIO::getChangeLog (const Package::Entry&  entry)
{
   Python::PyGILStateWrapper  gsw;

   QString  result;

   switch (entry.kind) {
   case Package::SUITE:
   case Package::PACKAGE:
      if (PyObject_HasAttrString(entry.obj, "hdr")) {
         // Must be a YumInstalledPackage.  Get info from the hdr

         PyObject*  namesList = getRpmHdrValue(entry.obj, "CHANGELOGNAME");
         if (namesList == NULL || namesList == Py_None) return QString();
         PyObject*  timesList = getRpmHdrValue(entry.obj, "CHANGELOGTIME");
         PyObject*  textsList = getRpmHdrValue(entry.obj, "CHANGELOGTEXT");

         PyObject*  namesIt = PyObject_GetIter(namesList);
         PyObject*  timesIt = (PyList_Check(timesList)
                               ? PyObject_GetIter(timesList)
                               : NULL);
         PyObject*  textsIt = PyObject_GetIter(textsList);
         for (;;) {
            PyObject*  nameItem = PyIter_Next(namesIt);
            if (nameItem == NULL) break;
            PyObject*  timeItem = (timesIt != NULL
                                   ? PyIter_Next(timesIt)
                                   : timesList);
            PyObject*  textItem = PyIter_Next(textsIt);

            QString  name = pyString(nameItem);
            QString  time = dateFormat(pyTime(timeItem));
            QString  text = pyString(textItem);

            if (!result.isEmpty()) {
               result.append("\n\n");
            }

            result.append("* ").append(time)
                  .append(' ').append(name)
                  .append('\n');
            result.append(text);

            Py_DECREF(textItem);
            Py_DECREF(timeItem);
            Py_DECREF(nameItem);
         }      

         Py_DECREF(textsIt);
         Py_XDECREF(timesIt);
         Py_DECREF(namesIt);

         Py_DECREF(textsList);
         Py_DECREF(timesList);
         Py_DECREF(namesList);
      } else if (PyObject_HasAttrString(entry.obj, "changelog")) {
         // Must be a YumAvailablePackage.  First make sure we've collected
         // the File Lists, and then get the files from returnFileEntries()
         bool  success;

         gsw.release();
         success = collectOtherData();
         gsw.ensure();

         if (success) {
            PyObject*  list;
            if (yumVersion() >= 20301) {
               list = pyMethodCall(entry.obj, "returnChangelog", "()");
            } else {
               list = PyObject_GetAttrString(entry.obj, "changelog");
            }
            if (list == NULL) {
               error(tr("Internal error: cannot find package's changelog"));
               return QString();
            }
            if (list == Py_None) {
               Py_DECREF(list);
               return QString();
            }

            PyObject*  it = PyObject_GetIter(list);
            for (;;) {
               PyObject*  item = PyIter_Next(it);
               if (item == NULL) break;

               QString  time = dateFormat(pyTime(PyTuple_GetItem(item, 0)));
               QString  name = pyString(PyTuple_GetItem(item, 1));
               QString  text = pyString(PyTuple_GetItem(item, 2));

               if (!result.isEmpty()) {
                  result.append("\n\n");
               }

               result.append("* ").append(time)
                     .append(' ').append(name)
                     .append('\n');
               result.append(text);

               Py_DECREF(item);
            }

            Py_DECREF(it);
            Py_DECREF(list);
         }
      }
      break;
   case Package::GROUP:
      // DO NOTHING
      break;
   }

   return result;
}

QString  YumIO::getRequiredGroups (const Package::Entry&  entry)
{
   return getGroupGrps(entry, "sub_groups");
}

QString  YumIO::getDefaultMetaPkgs (const Package::Entry&  entry)
{
   return getGroupGrps(entry, "default_metapkgs");
}

QString  YumIO::getOptionalMetaPkgs (const Package::Entry&  entry)
{
   return getGroupGrps(entry, "optional_metapkgs");
}

QString  YumIO::getMandatoryPkgs (const Package::Entry&  entry)
{
   return getGroupPkgs(entry, "mandatory_pkgs", "mandatory_packages");
}

QString  YumIO::getDefaultPkgs (const Package::Entry&  entry)
{
   return getGroupPkgs(entry, "default_pkgs", "default_packages");
}

QString  YumIO::getOptionalPkgs (const Package::Entry&  entry)
{
   return getGroupPkgs(entry, "optional_pkgs", "optional_packages");
}

QString  YumIO::getConditionalPkgs (const Package::Entry&  entry)
{
   if (hasDistribGroups()) {
      return getGroupPkgs(entry, "", "conditional_packages");
   } else {
      return QString();
   }
}

QString  YumIO::getYumVar (QString  name)
{
   PyObject*  conf   = PyObject_GetAttrString(yumIO, "conf");
   PyObject*  yumvar = PyObject_GetAttrString(conf, "yumvar");

   char*  nameStr = new char[name.length()+1];
   strcpy(nameStr, name.toStdString().c_str());

   PyObject*  resultObj = PyMapping_GetItemString(yumvar, nameStr);

   delete [] nameStr;

   Py_DECREF(yumvar);
   Py_DECREF(conf);

   return pyStringDec(resultObj);
}

class CollectFileListsThread : public QThread 
{
public:
   CollectFileListsThread(YumIO*       yumIO,
                          QEventLoop&  eventLoop)
      : yumIO(yumIO),
        eventLoop(eventLoop)
   {}
protected:
   YumIO*       yumIO;
   QEventLoop&  eventLoop;
public:
   bool         result;
protected:
   void run()
   {
      result = yumIO->collectFileListsReal();
      eventLoop.exit();
   }
};

bool  YumIO::collectFileLists ()
{
   if (collectedFileLists) return true;

   QEventLoop               eventLoop;
   CollectFileListsThread   thread(this, eventLoop);
   execWorkThread(eventLoop, thread);
   return thread.result;
}

bool  YumIO::collectFileListsReal ()
{
   blocker.block();

   Progress::User    pgUser(pg, totalProgressBar, itemProgressBar);
   YumIODialogsUser  yumIODialogsUser(dialogs, parent);

   log(Log::ACTION, 0, tr("Collecting File Lists"));

   repoProgress.kind = tr("file lists");
   repoProgress.act  = 0;

   repoProgress.parallel.begun   = false;
   repoProgress.parallel.setup   = false;
   repoProgress.parallel.meterMapping.clear();
   repoProgress.parallel.threads = NULL;
   repoProgress.parallel.retryMapping.clear();

   Python::PyGILStateWrapper  gsw;

   pg->startTotal(tr(PG_F_TOTAL_TEXT), PG_F_TOTAL);

   pg->startItem(tr(PG_F_REPOS_START_TEXT), PG_F_REPOS_START);

   pg->setupItemActs(PG_F_REPOS, repoProgress.repos.count() * RA_Count);
   {
#ifndef DEBUG_DISABLE_CAPTURERS
      YumIOCapturersUser  capturersUser(capturers, this);
#endif
      pyExceptStart();
      commExceptSetup();
      PyObject*  info = pyMethodCall(yumIO, "collectFileLists", "()");
      if (commExceptRaised(info, "collect filelists")) {
         pyExceptClear();
         pg->reset();
         return false;
      } else if (info == NULL) {
         error(tr("Internal error:"
                  " execution of yumIO.collectFileLists failed"));
         pyExceptClear();
         pg->reset();
         return false;
      }
      Py_DECREF(info);
      pyExceptClear();
   }
   pg->endItemActs();

   collectedFileLists = true;

   emit collectedExtraData();

   return true;
}

class CollectOtherDataThread : public QThread 
{
public:
   CollectOtherDataThread(YumIO*       yumIO,
                          QEventLoop&  eventLoop)
      : yumIO(yumIO),
        eventLoop(eventLoop)
   {}
protected:
   YumIO*       yumIO;
   QEventLoop&  eventLoop;
public:
   bool         result;
protected:
   void run()
   {
      result = yumIO->collectOtherDataReal();
      eventLoop.exit();
   }
};

bool  YumIO::collectOtherData ()
{
   if (collectedOtherData) return true;

   QEventLoop               eventLoop;
   CollectOtherDataThread   thread(this, eventLoop);
   execWorkThread(eventLoop, thread);
   return thread.result;
}

bool  YumIO::collectOtherDataReal ()
{
   blocker.block();

   Progress::User    pgUser(pg, totalProgressBar, itemProgressBar);
   YumIODialogsUser  yumIODialogsUser(dialogs, parent);

   log(Log::ACTION, 0, tr("Collecting Miscellaneous data"));

   repoProgress.kind = tr("miscellaneous data");
   repoProgress.act  = 0;

   repoProgress.parallel.begun   = false;
   repoProgress.parallel.setup   = false;
   repoProgress.parallel.meterMapping.clear();
   repoProgress.parallel.threads = NULL;
   repoProgress.parallel.retryMapping.clear();

   Python::PyGILStateWrapper  gsw;

   pg->startTotal(tr(PG_O_TOTAL_TEXT), PG_O_TOTAL);

   pg->startItem(tr(PG_O_REPOS_START_TEXT), PG_O_REPOS_START);

   pg->setupItemActs(PG_O_REPOS, repoProgress.repos.count() * RA_Count);
   {
#ifndef DEBUG_DISABLE_CAPTURERS
      YumIOCapturersUser  capturersUser(capturers, this);
#endif
      pyExceptStart();
      commExceptSetup();
      PyObject*  info = pyMethodCall(yumIO, "collectOtherData", "()");
      if (commExceptRaised(info, "collect miscellaneous data")) {
         pyExceptClear();
         pg->reset();
         return false;
      } else if (info == NULL) {
         error(tr("Internal error:"
                  " execution of yumIO.collectOtherData failed"));
         pyExceptClear();
         pg->reset();
         return false;
      }
      Py_DECREF(info);
      pyExceptClear();
   }
   pg->endItemActs();

   collectedOtherData = true;

   emit collectedExtraData();

   return true;
}

bool  YumIO::upToDateGroup (const Package::Entry&  entry)
{
   return (upToDateGroupGrps(entry, "sub_groups")
           && upToDateGroupPkgs(entry, "mandatory_pkgs", "mandatory_packages"));
}

bool  YumIO::upToDateGroupGrps (const Package::Entry&  entry,
                                const char*            attr_22)
{
   char*  copy = new char[strlen(attr_22)+1];
   strcpy(copy, attr_22);
   bool  result = upToDateGroupGrps(entry, copy);
   delete [] copy;
   return result;
}

bool  YumIO::upToDateGroupGrps (const Package::Entry&  entry,
                                char*                  attr_22)
{
   if (hasDistribGroups()) {
      // DO NOTHING
   } else {
      PyObject*  seq = getGroupItem_22(entry, attr_22);
      if (seq == NULL) {
         return true;
      }
      if (PySequence_Length(seq) > 0) {
         QStringList  values;

         PyObject*  it = PyObject_GetIter(seq);
         for (;;) {
            PyObject*  grpId = PyIter_Next(it);
            if (grpId == NULL) break;
            EntryKey  key(true, pyString(grpId), QString());
            if (hash.contains(key)) {
               if (!upToDateGroup(*hash[key])) {
                  Py_DECREF(seq);
                  return false;
               }
            }
            Py_DECREF(grpId);
         }

         Py_DECREF(it);
      }
      Py_DECREF(seq);
   }

   return true;
}

bool  YumIO::upToDateGroupPkgs (const Package::Entry&  entry,
                                const char*            attr_22,
                                const char*            attr_25)
{
   char*  attr_22_copy = new char[strlen(attr_22)+1];
   char*  attr_25_copy = new char[strlen(attr_25)+1];
   strcpy(attr_22_copy, attr_22);
   strcpy(attr_25_copy, attr_25);
   bool  result = upToDateGroupPkgs(entry, attr_22_copy, attr_25_copy);
   delete [] attr_25_copy;
   delete [] attr_22_copy;
   return result;
}

bool  YumIO::upToDateGroupPkgs (const Package::Entry&  entry,
                                char*                  attr_22,
                                char*                  attr_25)
{
   if (hasDistribGroups()) {
      PyObject*  dict = getGroupItem_25(entry, attr_25);
      if (dict == NULL) {
         return true;
      }

      if (PyDict_Size(dict) > 0) {
         Py_ssize_t  pos = 0;
         for (;;) {
            PyObject*  pkgName;
            PyObject*  value;
            if (!PyDict_Next(dict, &pos, &pkgName, &value)) break;

            Package::Entry*  pkg = findPackage(pyString(pkgName));
            if (pkg != NULL) {
               if (pkg->state == Package::OUT_OF_DATE) {
                  Py_DECREF(dict);
                  return false;
               }
            }
         }
      }

      Py_DECREF(dict);
   } else {
      PyObject*  seq = getGroupItem_22(entry, attr_22);
      if (seq == NULL) {
         return true;
      }

      if (PySequence_Length(seq) > 0) {
         PyObject*  it = PyObject_GetIter(seq);
         for (;;) {
            PyObject*  pkgName = PyIter_Next(it);
            if (pkgName == NULL) break;
            Package::Entry*  pkg = findPackage(pyString(pkgName));
            if (pkg != NULL) {
               if (pkg->state == Package::OUT_OF_DATE) {
                  Py_DECREF(seq);
                  return false;
               }
            }
            Py_DECREF(pkgName);
         }

         Py_DECREF(it);
      }
      Py_DECREF(seq);
   }

   return true;
}

QString  YumIO::getPrco (PyObject*    pkg,
                         const char*  hdrKey,
                         const char*  prcoKey)
{
   char*  procKeyCopy = new char[strlen(prcoKey)+1];
   strcpy(procKeyCopy, prcoKey);
   QString  result = getPrco(pkg, hdrKey, procKeyCopy);
   delete [] procKeyCopy;
   return result;
}

QString  YumIO::getPrco (PyObject*    pkg,
                         const char*  hdrKey,
                         char*        prcoKey)
{
   Python::PyGILStateWrapper  gsw;

   QStringList  values;

   if (PyObject_HasAttrString(pkg, "hdr")) {
      // Must be a YumInstalledPackage.  Get the prco from the hdr
      char  nameKey[1024];
      char  flagsKey[1024];
      char  versionKey[1024];
      sprintf(nameKey,    "%sNAME",    hdrKey);
      sprintf(flagsKey,   "%sFLAGS",   hdrKey);
      sprintf(versionKey, "%sVERSION", hdrKey);

      PyObject*  nameList    = getRpmHdrValue(pkg, nameKey);
      if (nameList == NULL || nameList == Py_None) return QString();
      PyObject*  flagsList   = getRpmHdrValue(pkg, flagsKey);
      PyObject*  versionList = getRpmHdrValue(pkg, versionKey);

      PyObject*  nameIt    = PyObject_GetIter(nameList);
      PyObject*  flagsIt   = (flagsList != NULL ? PyObject_GetIter(flagsList)
                                                : NULL);
      PyObject*  versionIt = (flagsList != NULL ? PyObject_GetIter(versionList)
                                                : NULL);
      for (;;) {
         PyObject*  nameItem    = PyIter_Next(nameIt);
         if (nameItem == NULL) break;
         PyObject*  flagsItem   = (flagsIt != NULL   ? PyIter_Next(flagsIt)
                                                     : NULL);
         PyObject*  versionItem = (versionIt != NULL ? PyIter_Next(versionIt)
                                                     : NULL);

         QString  value = pyString(nameItem);

         if (   flagsItem   != NULL && flagsItem   != Py_None
             && versionItem != NULL && versionItem != Py_None) {
            unsigned long  flags = PyInt_AsLong(flagsItem);
            QString        flagsStr;
            if (flags & RPMSENSE_LESS)    flagsStr.append('<');
            if (flags & RPMSENSE_GREATER) flagsStr.append('>');
            if (flags & RPMSENSE_EQUAL)   flagsStr.append('=');
            value.append(' ').append(flagsStr);
            value.append(' ').append(pyString(versionItem));
         }

         values.append(value);

         Py_XDECREF(versionItem);
         Py_XDECREF(flagsItem);
         Py_DECREF(nameItem);
      }      

      Py_XDECREF(versionIt);
      Py_XDECREF(flagsIt);
      Py_DECREF(nameIt);

      Py_DECREF(versionList);
      Py_DECREF(flagsList);
      Py_DECREF(nameList);
   } else if (PyObject_HasAttrString(pkg, "prco")) {
      // Must be a YumAvailablePackage.  
      //
      // If the package has a returnPrco function, as is the case with newer
      // versions of yum (e.g. 2.4.2), then call it.  This is necessary even
      // if they have a prco attribute, because it is filled in lazily.
      //
      // However, for older versions of yum (e.g. 2.2.2), there is no
      // returnPrco function, so use the prco attribute directly.  Note that
      // returnSimple won't return the prco attribute.
      PyObject*  list;
      if (PyObject_HasAttrString(pkg, "returnPrco")) {
         list = pyMethodCall(pkg, "returnPrco", "(s)", prcoKey);
      } else {
         PyObject*  prco = PyObject_GetAttrString(pkg, "prco");
         list = PyDict_GetItemString(prco, prcoKey);
         Py_DECREF(prco);
      }
      if (list == NULL) {
         error(tr("Internal error:"
                  " cannot find package's prco.%1").arg(prcoKey));
         return QString();
      }

      PyObject*  it = PyObject_GetIter(list);
      for (;;) {
         PyObject*  item = PyIter_Next(it);
         if (item == NULL) break;

         PyObject*  name  = PyTuple_GetItem(item, 0);
         PyObject*  flags = PyTuple_GetItem(item, 1);
         PyObject*  evr   = PyTuple_GetItem(item, 2);

         QString  value = pyString(name);
         
         if (   flags != NULL && flags != Py_None
             && evr   != NULL && evr   != Py_None) {
            const char*  flagsCode = Python::pyString(flags);
            const char*  flagsStr  = YUM_DECODE_PRCO_FLAGS(flagsCode);
            PyObject*    epoch     = PyTuple_GetItem(evr, 0);
            PyObject*    ver       = PyTuple_GetItem(evr, 1);
            PyObject*    rel       = PyTuple_GetItem(evr, 2);
            if (ver != NULL && ver != Py_None) {
               value.append(' ').append(flagsStr).append(' ');
               if (epoch != NULL && epoch != Py_None) {
                  char*  epochStr = Python::pyString(epoch);
                  if (strcmp(epochStr, "0") != 0) {
                     value.append(epochStr).append(':');
                  }
               }
               value.append(pyString(ver));
               if (rel != NULL && rel != Py_None) {
                  value.append('-').append(pyString(rel));
               }
            }
         }

         values.append(value);

         Py_DECREF(item);
      }

      Py_DECREF(it);
      if (PyObject_HasAttrString(pkg, "returnPrco")) {
         Py_DECREF(list);
      }
   }

   return makeSorted(values);
}

#if 0
QString  YumIO::getRpm(PyObject*  pkg,
                       char*      rpmKey)
{
   Python::PyGILStateWrapper  gsw;

   if (PyObject_HasAttrString(pkg, "hdr")) {
      return pyStringXDec(getRpmHdrValue(pkg, rpmKey));
   } else {
      return QString();
   }
}
#endif

QString  YumIO::getRpmTime(PyObject*    pkg,
                           const char*  rpmKey)
{
   Python::PyGILStateWrapper  gsw;

   if (PyObject_HasAttrString(pkg, "hdr")) {
      return timeFormat(pyTimeDec(getRpmHdrValue(pkg, rpmKey)));
   } else {
      return QString();
   }
}

QString  YumIO::getRpmList(PyObject*    pkg,
                           const char*  rpmKey)
{
   Python::PyGILStateWrapper  gsw;

   if (PyObject_HasAttrString(pkg, "hdr")) {
      QString  value;

      PyObject*  list = getRpmHdrValue(pkg, rpmKey);
      if (list != NULL && list != Py_None) {
         PyObject*  it = PyObject_GetIter(list);
         for (;;) {
            PyObject*  item = PyIter_Next(it);
            if (item == NULL) break;
            if (!value.isEmpty()) {
               value.append('\n');
            }
            value.append(pyString(item));
            Py_DECREF(item);
         }
         Py_DECREF(it);
         Py_DECREF(list);
      }

      return value;
   } else {
      return QString();
   }
}

QString  YumIO::getGroupGrps (const Package::Entry&  entry,
                              const char*            attr_22)
{
   char*  copy = new char[strlen(attr_22)+1];
   strcpy(copy, attr_22);
   QString  result = getGroupGrps(entry, copy);
   delete [] copy;
   return result;
}

QString  YumIO::getGroupGrps (const Package::Entry&  entry,
                              char*                  attr_22)
{
   Python::PyGILStateWrapper  gsw;

   if (hasDistribGroups()) {
      return QString();
   } else {
      PyObject*  seq = getGroupItem_22(entry, attr_22);
      if (seq == NULL) {
         return QString();
      }

      if (PySequence_Length(seq) > 0) {
         QStringList  values;

         PyObject*  it = PyObject_GetIter(seq);
         for (;;) {
            PyObject*  grpId = PyIter_Next(it);
            if (grpId == NULL) break;
            values.append(pyString(grpId));
            Py_DECREF(grpId);
         }

         Py_DECREF(it);

         Py_DECREF(seq);
         return makeSorted(values);
      } else {
         Py_DECREF(seq);
         return QString();
      }
   }
}

QString  YumIO::getGroupPkgs (const Package::Entry&  entry,
                              const char*            attr_22,
                              const char*            attr_25)
{
   char*  attr_22_copy = new char[strlen(attr_22)+1];
   char*  attr_25_copy = new char[strlen(attr_25)+1];
   strcpy(attr_22_copy, attr_22);
   strcpy(attr_25_copy, attr_25);
   QString  result = getGroupPkgs(entry, attr_22_copy, attr_25_copy);
   delete [] attr_25_copy;
   delete [] attr_22_copy;
   return result;
}

QString  YumIO::getGroupPkgs (const Package::Entry&  entry,
                              char*                  attr_22,
                              char*                  attr_25)
{
   Python::PyGILStateWrapper  gsw;

   if (hasDistribGroups()) {
      PyObject*  dict = getGroupItem_25(entry, attr_25);
      if (dict == NULL) {
         return QString();
      }

      if (PyDict_Size(dict) > 0) {
         QStringList  values;

         Py_ssize_t  pos = 0;
         for (;;) {
            PyObject*  pkgName;
            PyObject*  value;
            if (!PyDict_Next(dict, &pos, &pkgName, &value)) break;

            values.append(pyString(pkgName));
         }

         Py_DECREF(dict);
         return makeSorted(values);
      } else {
         Py_DECREF(dict);
         return QString();
      }
   } else {
      PyObject*  seq = getGroupItem_22(entry, attr_22);
      if (seq == NULL) {
         return QString();
      }

      if (PySequence_Length(seq) > 0) {
         QStringList  values;

         PyObject*  it = PyObject_GetIter(seq);
         for (;;) {
            PyObject*  pkgName = PyIter_Next(it);
            if (pkgName == NULL) break;
            values.append(pyString(pkgName));
            Py_DECREF(pkgName);
         }

         Py_DECREF(it);

         Py_DECREF(seq);
         return makeSorted(values);
      } else {
         Py_DECREF(seq);
         return QString();
      }
   }
}

QString  YumIO::makeSorted (QStringList&  list)
{
   qSort(list.begin(), list.end());

   QString  result;

   QStringList::const_iterator  it = list.begin();
   if (it != list.end()) {
      result.append(*it); ++it;
   }
   for (; it != list.end(); ++it) {
      result.append("\n").append(*it);
   }

   return result;
}

QString  YumIO::makeSortedUniq (QStringList&  list)
{
   qSort(list.begin(), list.end());

   QString  result;

   QStringList::const_iterator  it   = list.begin();
   QStringList::const_iterator  prev = list.begin();
   if (it != list.end()) {
      result.append(*it); prev = it; ++it;
   }
   for (; it != list.end(); prev = it, ++it) {
      if (*it != *prev) {
         result.append("\n").append(*it);
      }
   }

   return result;
}

PyObject*  YumIO::getRpmHdrValue(PyObject*    pkg,
                                 const char*  rpmKey)
{
   if (!PyObject_HasAttrString(pkg, "hdr")) {
      error(tr("getRpmHdrValue called for package with no hdr"
               " (i.e. not installed)"));
      return NULL;
   }

   PyObject*  hdr = PyObject_GetAttrString(pkg, "hdr");

   char  key[1024];
   sprintf(key, "RPMTAG_%s", rpmKey);

   PyObject*  tag = PyDict_GetItemString(rpm_dict, key);
   if (tag == NULL) {
      error(tr("Internal error: cannot find rpm.%1").arg(key));
      return NULL;
   }

   PyObject*  result = PyObject_GetItem(hdr, tag);
   Py_DECREF(hdr);
   return result;
}

class ApplyThread : public QThread 
{
public:
   ApplyThread(YumIO*       yumIO,
               QEventLoop&  eventLoop)
      : yumIO(yumIO),
        eventLoop(eventLoop)
   {}
protected:
   YumIO*       yumIO;
   QEventLoop&  eventLoop;
public:
   bool         result;
protected:
   void run()
   {
      result = yumIO->applyReal();
      eventLoop.exit();
   }
};

bool YumIO::apply () 
{
   QEventLoop   eventLoop;
   ApplyThread  thread(this, eventLoop);
   execWorkThread(eventLoop, thread);
   return thread.result;
}

bool YumIO::applyReal () 
{
   blocker.block();

   Progress::User    pgUser(pg, totalProgressBar, itemProgressBar);
   YumIODialogsUser  yumIODialogsUser(dialogs, parent);

   log(Log::ACTION, 0, tr("Applying"));

   Python::PyGILStateWrapper  gsw;

   pg->startTotal(tr(PG_D_TOTAL_TEXT), PG_D_TOTAL);

   pg->startItem(tr(PG_D_INITIALIZE_TEXT), PG_D_INITIALIZE);

   ts.descr               = QString();
   ts.confirmed           = false;
   ts.act                 = 0;
   ts.dep.totalTs         = 0;
   ts.dep.totalGrabs      = 0;
   ts.run.nominalGrabSize = 0;
   ts.run.totalGrabSize   = 0;
   ts.run.totalTsSize     = 0;
   ts.run.lastTs          = 0;
   ts.run.deltaRebuilt    = 0;
   ts.parallel.begun      = false;
   ts.parallel.meterMapping.clear();

   PyObject*  installs = PyList_New(0);
   PyObject*  updates  = PyList_New(0);
   PyObject*  removes  = PyList_New(0);

   {
      Package::EntryList::const_iterator  it = list.begin();
      for (; it != list.end(); ++it) {
         Package::Entry*  entry = *it;
         entry->apply.action = entry->action;
         entry->apply.reason = entry;
      }
   }

   bool  status = applyGroups();
   if (!status) {
      return false;
   }

   // I'm not bothering to Py_INCREF and Py_DECREF the packages themselves,
   // because I already am holding a reference for them because they're in
   // the entries list.  That's good enough.

   {
      Package::EntryList::const_iterator  it = list.begin();
      for (; it != list.end(); ++it) {
         Package::Entry*  entry = *it;

         if (entry->kind != Package::GROUP) {
            switch (entry->apply.action) {
            case Package::UPDATE:
               {
                  if (entry->next == NULL) {
                     error(tr("Unable to find installed version for package"
                              " update %1").arg(entry->name));
                     return false;
                  }
                  PyObject*  pair = PyList_New(0);
                  status = PyList_Append(pair, entry->obj);
                  if (status != 0) {
                     error(tr("Unable to add new package to pair for updated"
                              " package %1").arg(entry->name));
                  }
                  status = PyList_Append(pair, entry->next->obj);
                  if (status != 0) {
                     error(tr("Unable to add old package to pair for updated"
                              " package %1").arg(entry->name));
                  }
                  status = PyList_Append(updates, pair);
                  if (status != 0) {
                     error(tr("Unable to add package %1 to updates"
                              " list").arg(entry->name));
                  }
                  ts.dep.totalTs += 2;
                  if (!QFile::exists(getLocalHdr(entry->obj))) {
                     ts.dep.totalGrabs++;
                  }
               }
               break;
            case Package::INSTALL:
               {
                  int  status = PyList_Append(installs, entry->obj);
                  if (status != 0) {
                     error(tr("Unable to add package %1 to installs"
                              " list").arg(entry->name));
                  }
                  ts.dep.totalTs++;
                  if (!QFile::exists(getLocalHdr(entry->obj))) {
                     ts.dep.totalGrabs++;
                  }
               }
               break;
            case Package::REMOVE:
               {
                  Package::Entry*  e = entry;
                  if (e->next != NULL) {
                     e = e->next;
                  }
                  int  status = PyList_Append(removes, e->obj);
                  if (status != 0) {
                     error(tr("Unable to add package %1 to removes"
                              " list").arg(entry->name));
                  }
                  ts.dep.totalTs++;
               }
               break;
            case Package::NONE:
               // DO NOTHING
               break;
            }
         }
      }
   }

   // Assume that all headers are roughly the same size and use number of
   // grabs as the number of acts.
   unsigned int  grabActs = PG_D_GRAB(ts.dep.totalGrabs);
   ts.dep.depSolveActs      = PG_D_DEPSOLVE_ESTIMATE(ts.dep.totalTs);
   ts.dep.totalDepSolveActs = ts.dep.depSolveActs;
   unsigned int  numActs = grabActs + ts.dep.totalDepSolveActs;
   pg->setupItemActs(PG_D_DEPSOLVE, numActs);

   bool  success = reapChildren();
   if (!success) {
      pg->reset();
      return false;
   }

   {
#ifndef DEBUG_DISABLE_CAPTURERS
      YumIOCapturersUser  capturersUser(capturers, this);
#endif
      sigchldDisable();
      pyExceptStart();
      commExceptSetup();
      PyObject*  applyStatus = pyMethodCall(yumIO, "apply", "(O,O,O)", 
                                            installs, updates, removes);
      if (commExceptRaised(applyStatus, "apply changes")) {
         pyExceptClear();

         // Set ts.confirmed so we'll return false, and let the rest of
         // this routine do its clean up.
         ts.confirmed = false;

         // At least one major purpose of this is to call self.closeRpmDB ->
         // self._ts.close -> self._ts.ts.closeDB -> rpmts_CloseDB -> 
         // rpmdbClose -> rpmsqEnable, which prevents rpmdbCheckSignals
         // from calling exit(1) if a SIGINT caused apply to fail.
         PyObject*  applyFailCleanupStatus;
         applyFailCleanupStatus = pyMethodCall(yumIO, "applyFailCleanup", "()");
         Py_XDECREF(applyFailCleanupStatus);
      } else {
         Py_XDECREF(applyStatus);
         pyExceptClear();
      }

      sigchldRestore();
   }

   PyObject*  flushStatus;
   flushStatus = pyMethodCall(yumIO, "flush", "()");
   Py_XDECREF(flushStatus);

   Py_DECREF(removes);
   Py_DECREF(updates);
   Py_DECREF(installs);

   pg->reset();

   return ts.confirmed;
}

bool  YumIO::cleanHeaders ()
{
   log(Log::ACTION, 0, tr("Cleaning Headers"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanHeaders", "()");

   bool  result = (PyInt_AsLong(PyTuple_GetItem(pair, 0)) == 0);
   Py_XDECREF(pair);
   return result;
}

bool  YumIO::cleanPackages ()
{
   log(Log::ACTION, 0, tr("Cleaning Packages"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanPackages", "()");

   bool  result = (PyInt_AsLong(PyTuple_GetItem(pair, 0)) == 0);
   Py_XDECREF(pair);
   return result;
}

bool  YumIO::cleanXmlMetadata ()
{
   log(Log::ACTION, 0, tr("Cleaning Xml Metadata"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanMetadata", "()");

   bool  result = (PyInt_AsLong(PyTuple_GetItem(pair, 0)) == 0);
   Py_XDECREF(pair);
   return result;
}

bool  YumIO::cleanPickledCache ()
{
   log(Log::ACTION, 0, tr("Cleaning Pickled Cache"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanPickles", "()");

   bool  result = (PyInt_AsLong(PyTuple_GetItem(pair, 0)) == 0);
   Py_XDECREF(pair);
   return result;
}

bool  YumIO::cleanSqliteCache ()
{
   log(Log::ACTION, 0, tr("Cleaning Sqlite Cache"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanSqlite", "()");

   bool  result = (PyInt_AsLong(PyTuple_GetItem(pair, 0)) == 0);
   Py_XDECREF(pair);
   return result;
}

bool  YumIO::cleanAll ()
{
   log(Log::ACTION, 0, tr("Cleaning All"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  hdrPair  = pyMethodCall(yumIO, "cleanHeaders", "()");
   PyObject*  pkgPair  = pyMethodCall(yumIO, "cleanPackages", "()");
   PyObject*  metaPair = pyMethodCall(yumIO, "cleanMetadata", "()");

   long  hdrValue  = PyInt_AsLong(PyTuple_GetItem(hdrPair, 0));
   long  pkgValue  = PyInt_AsLong(PyTuple_GetItem(pkgPair, 0));
   long  metaValue = PyInt_AsLong(PyTuple_GetItem(metaPair, 0));
   long  pickValue = 0;
   long  sqlValue  = 0;
   
   Py_XDECREF(hdrPair);
   Py_XDECREF(pkgPair);
   Py_XDECREF(metaPair);

   if (hasPickledCache()) {
      PyObject*  pickPair = pyMethodCall(yumIO, "cleanPickles", "()");
      pickValue = PyInt_AsLong(PyTuple_GetItem(pickPair, 0));
      Py_XDECREF(pickPair);
   }
   if (hasSqliteCache()) {
      PyObject*  sqlPair = pyMethodCall(yumIO, "cleanSqlite", "()");
      sqlValue = PyInt_AsLong(PyTuple_GetItem(sqlPair, 0));
      Py_XDECREF(sqlPair);
   }

   return (hdrValue == 0
           && pkgValue == 0
           && metaValue == 0
           && pickValue == 0
           && sqlValue == 0);
}

/*
** yumIO callback functions from Python
*/

PyObject*  YumIO::stdoutWrite(PyObject*  self, PyObject*  args)
{ 
   return write(self, args, Log::STDOUT, 0, wr.stdoutBuffer);
}

PyObject*  YumIO::stderrWrite(PyObject*  self, PyObject*  args)
{ 
   return write(self, args, wr.stderrKind, wr.stderrValue, wr.stderrBuffer);
}

PyObject*  YumIO::write(PyObject*     self, 
                        PyObject*     args,
                        Log::Kind     kind,
                        unsigned int  value,
                        QString&      buffer)
{
   char*  content;
   
   if (!PyArg_ParseTuple(args, "s", &content)) return NULL;

   write(content, kind, value, buffer);

   Py_INCREF(Py_None);
   return Py_None;
}

void  YumIO::write (QString       content,
                    Log::Kind     kind,
                    unsigned int  value,
                    QString&      buffer)
{
   buffer.append(content);

   QStringList  lines = buffer.split("\n");

   QStringList::const_iterator  it   = lines.begin();
   QStringList::const_iterator  last = lines.end() - 1;

   if (it != last) {
      logEntries.lock();

      int i = 0;
      for (; it != last; ++it, ++i) {
         logEntries.append(Log::Entry(kind, value, *it));
      }

      logEntries.unlock();
      logEntries.emitAppended();
   }

   // The last line in the list is, by definition, unfinished.  (Note that
   // a very common case would be when it's the empty string.)  Sit on it
   // until further text comes along to finish it with a newline.
   buffer = *it;
}

void  YumIO::writeStdout (QString  content)
{
   write(content, Log::STDOUT, 0, wr.stdoutBuffer);
}

void  YumIO::writeStderr (QString  content)
{
   write(content, Log::ERROR, 0, wr.stderrBuffer);
}

PyObject*  YumIO::log(PyObject*  self, PyObject*  args)
{ 
   long   value;
   char*  msg;
   
   if (!PyArg_ParseTuple(args, "ls", &value, &msg)) return NULL;

   log(Log::NORMAL, value, msg);

   if (strncmp(msg,
               APPLY_DEPSOLVE_PREFIX,
               STRLEN(APPLY_DEPSOLVE_PREFIX)) == 0) {
      // I hate having to know this string, but there is no callback
      // for a package added to the transaction by depSolve.  (There's
      // one for the Requires: string that caused it, but that's not
      // good enough.)
      dsTsInfo(msg + STRLEN(APPLY_DEPSOLVE_PREFIX));
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::errorlog(PyObject*  self, PyObject*  args)
{ 
   long   value;
   char*  msg;
   
   if (!PyArg_ParseTuple(args, "ls", &value, &msg)) return NULL;

   log(Log::ERROR, value, msg);

   Py_INCREF(Py_None);
   return Py_None;
}

void  YumIO::collectStart (void*    meterObj,
                           QString  name,
                           RepoAct  act,
                           long     current,
                           long     total)
{
   if (hasMultiProgress()) {
      // In yum 3.4.3, RedHat patched changes from the GIT tree which
      // support a new parallelized ordering of events which results in
      // this ordering:
      //    Repo 1: act RA_Grab_Metalink
      //    Repo 1: act RA_Grab_RepoMD
      //    Repo 2: act RA_Grab_Metalink
      //    Repo 2: act RA_Grab_RepoMD
      //    ...
      //    Repo n: act RA_Grab_Metalink
      //    Repo n: act RA_Grab_RepoMD
      //    ----------------------------
      //    Repo 1: act RA_Grab_Group    |
      //    Repo 1: act RA_Grab          |
      //    Repo 2: act RA_Grab_Group    |
      //    Repo 2: act RA_Grab           > (in parallel)
      //    ..                           |
      //    Repo n: act RA_Grab_Group    |
      //    Repo n: act RA_Grab          |
      //    ----------------------------
      //    Repo 1: act RA_Convert
      //    Repo 1: act RA_Read
      //    Repo 2: act RA_Convert
      //    Repo 2: act RA_Read
      //    ...
      //    Repo n: act RA_Convert
      //    Repo n: act RA_Read
      if (act <= RA_Grab_RepoMD) {
         repoProgress.act = (repoProgress.act / 2 * 2 + (int)act + 1);
      } else if (act <= RA_Grab) {
         int  passTwoStart = repoProgress.repos.count() * 2;
         if (repoProgress.act < passTwoStart) {
            repoProgress.act = passTwoStart;
         }
         repoProgress.act = ((repoProgress.act - passTwoStart) / 2 * 2
                             + (int)act - 2 + 1) + passTwoStart;
      } else {
         int  passThreeStart = repoProgress.repos.count() * 4;
         if (repoProgress.act < passThreeStart) {
            repoProgress.act = passThreeStart;
         }
         repoProgress.act = ((repoProgress.act - passThreeStart) / 2 * 2
                             + (int)act - 4 + 1) + passThreeStart;
      }
   } else if (yumVersion() >= 30107) {
      // In versions of yum starting with 3.1.7, loadRepoXML gets called
      // lazily the first time the repomd.xml is needed.  As a result, acts
      // are arranged thusly:
      //    Repo 1: act 1
      //    Repo 1: act ...
      //    Repo 1: act RA_Read
      //    Repo 2: act 1
      //    Repo 2: act ...
      //    Repo 2: act RA_Read
      //    ...
      // If any previous act for the repo is skipped, this code should move
      // to the correct act number for the given repo.  This does still rely
      // on the fact that an RA_Read always will happen for every repo to
      // move the act to the final act for the given repo.  This is the only
      // way subsequent acts will end up with numbers appropriate for a
      // subsequent repo.
      repoProgress.act = (repoProgress.act / RA_Count * RA_Count
                          + (int)act + 1);
   } else {
      // In pre-3.1.7 versions of yum, loadRepoXML gets called manually
      // before calling doSackSetup, and this results in the following odd
      // order for acts.  Acts are arranged thusly:
      //    Repo 1: RA_Grab_Metalink
      //    Repo 1: RA_Grab_RepoMD
      //    Repo 2: RA_Grab_Metalink
      //    Repo 2: RA_Grab_RepoMD
      //    ...
      //    Repo n: RA_Grab_Metalink
      //    Repo n: RA_Grab_RepoMD
      //    ------------------------
      //    Repo 1: RA_Grab
      //    Repo 1: RA_Convert
      //    Repo 1: RA_Read
      //    Repo 2: RA_Grab
      //    Repo 2: RA_Convert
      //    Repo 2: RA_Read
      //    ...
      //    Repo n: RA_Grab
      //    Repo n: RA_Convert
      //    Repo n: RA_Read
      // That is, things happen in two passes.  The first pass walks all
      // the repos and does the first two acts for each.  The second pass
      // walks all the repos and does the last 3 acts for each.
      //
      // In the first pass, if the RA_Grab_Metalink is skipped, but I see the
      // RA_Grab_RepoMD, I'll jump ahead by two to catch up.  However if I
      // get no callback for a particular repo, I simply won't know what's
      // up, so I may fall pretty far behind.  However, once I see the first
      // callback for anything in the second pass, I immediately will jump to
      // the appropriate spot and resume from there.
      //
      // In the second pass, if any previous act for the repo is skipped,
      // this code should move to the correct act number for the given repo.
      // This does still rely on the fact that an RA_Read always will happen
      // for every repo to move the act to the final act for the given repo.
      // This is the only way subsequent acts will end up with numbers
      // appropriate for a subsequent repo.
      if (act <= RA_Grab_RepoMD) {
         repoProgress.act = (repoProgress.act / 2 * 2 + (int)act + 1);
      } else {
         int  passTwoStart = repoProgress.repos.count() * 2;
         if (repoProgress.act < passTwoStart) {
            repoProgress.act = passTwoStart;
         }
         repoProgress.act   = ((repoProgress.act - passTwoStart) / 3 * 3
                               + (int)act - 2 + 1) + passTwoStart;
      }
   }

   repoProgress.total = total;
   repoProgress.name  = name;

#if defined(DEBUG_COLLECT)
   printf("collectStart %p %-11s name=%s, act=%ld/%ld, progress=%d/%d\n",
          meterObj, RA_IMAGE(act), name.toStdString().c_str(), current, total,
          repoProgress.act, repoProgress.repos.count() * RA_Count);
#endif

   QString  activity;
   switch (act) {
   case RA_Grab_Metalink:
      if (name.isEmpty()) {
         activity = tr("Downloading repository mirror list");
      } else {
         activity = tr("Downloading repository mirror list for %2").arg(name);
      }
      break;
   case RA_Grab_RepoMD:
      if (name.isEmpty()) {
         activity = tr("Downloading repository repomd");
      } else {
         activity = tr("Downloading repository repomd for %2").arg(name);
      }
      break;
   case RA_Grab_Group:
      if (name.isEmpty()) {
         activity = tr("Downloading repository groups");
      } else {
         activity = tr("Downloading repository groups for %2").arg(name);
      }
      break;
   case RA_Grab:
      if (name.isEmpty()) {
         activity = tr("Downloading repository %1")
                       .arg(repoProgress.kind);
      } else {
         activity = tr("Downloading repository %1 for %2")
                       .arg(repoProgress.kind).arg(name);
      }
      break;
   case RA_Convert:
      if (name == "") {
         activity = tr("Converting repository");
      } else {
         activity = tr("Converting repository %2").arg(name);
      }
      break;
   case RA_Read:
      if (name == "") {
         activity = tr("Reading repository %1")
                       .arg(repoProgress.kind);
      } else {
         activity = tr("Reading repository %1 for %2")
                       .arg(repoProgress.kind).arg(name);
      }
      break;
   default:
      error("Internal error: invalid collect act\n");
   }

   if (repoProgress.parallel.begun) {
      if (!repoProgress.parallel.setup) {
         QString          activity("Unknown");
         Progress::Steps  firstAct = 0;
         Progress::Steps  tailAct  = 1;
         if (hasMultiProgress()) {
            if (act == RA_Grab_Group || act == RA_Grab) {
               activity = QString("Downloading repository contents for"
                                  " %1 repositories")
                             .arg(repoProgress.parallel.files);
               // The RA_Grab_Group/RA_Grab downloads are partitioned into a
               // parallel pass with known sizes, and another parallel pass
               // with unknown sizes.  (See repos.py, 
               // repos.RepoStorage.retrieveAllMD.)
               if (repoProgress.parallel.size != 0) {
                  firstAct = repoProgress.repos.count() * 2;
                  tailAct  = repoProgress.repos.count() * 3;
               } else {
                  firstAct = repoProgress.repos.count() * 3;
                  tailAct  = repoProgress.repos.count() * 4;
               }
            } else {
               error(tr("Internal error: Unexpected parallel grab type:\n"
                        "%1").arg(name));
            }
         } else {
            error(tr("Internal error: Unexpected parallel grab"
                     " with no parallel support:\n%1").arg(name));
         }
         pg->setupItemActsThreads(activity,
                                  firstAct,
                                  tailAct,
                                  repoProgress.parallel.files,
                                  repoProgress.parallel.size);
         repoProgress.parallel.setup = true;
      }

      Progress::Threads  thread;
      // Determine if this is a retry.  If so, reuse the same thread number
      // as before.  If not, assign a new thread number.
      RetryMapping::const_iterator  it
         = repoProgress.parallel.retryMapping.find(ThreadId(name, act));
      if (it != repoProgress.parallel.retryMapping.end()) {
         thread = it.value();
      } else {
         thread = repoProgress.parallel.threadsStarted++;
      }
#if defined(DEBUG_COLLECT)
      printf("collectStart %p %-11s name=%s thread=%d\n",
             meterObj, RA_IMAGE(act), name.toStdString().c_str(), thread);
#endif
      if (thread > repoProgress.parallel.files) {
         error(tr("Internal error: Start: unexpected thread %1"
                  " > total files %2\n")
                  .arg(thread).arg(repoProgress.parallel.files));
      } else {
         repoProgress.parallel.meterMapping[meterObj] = thread;

         ThreadId&  tid = repoProgress.parallel.threads[thread];
         tid.name = name;
         tid.act  = act;
         repoProgress.parallel.retryMapping[tid] = thread;
         pg->startItemActThread(activity, thread, total);
         if (current != 0) {
            pg->setItemActsThreadValue(thread, current);
         }
      }
   } else {
      pg->startItemAct(activity, total, repoProgress.act);
      if (current != 0) {
         pg->setItemValue(current);
      }
   }
}

void  YumIO::collectUpdate (void*    meterObj,
                            long     current)
{
#if defined(DEBUG_COLLECT_DETAIL)
   printf("collectUpdate %p (progress=%ld)\n", meterObj, current);
#endif
   if (repoProgress.parallel.begun) {
      Progress::Threads  thread = repoProgress.parallel.meterMapping[meterObj];
      if (thread > repoProgress.parallel.files) {
         error(tr("Internal error: Update: unexpected thread %1"
                  " > total files %2\n")
                  .arg(thread).arg(repoProgress.parallel.files));
      } else {
         pg->setItemActsThreadValue(thread, current);
      }
   } else {
      pg->setItemValue(current);
   }
}

void YumIO::collectEnd (void*  meterObj,
                        long   current)
{
#if defined(DEBUG_COLLECT)
   printf("collectEnd %p (progress=%ld)\n", meterObj, current);
#endif

   if (repoProgress.parallel.begun) {
      Progress::Threads  thread = repoProgress.parallel.meterMapping[meterObj];

      if (thread > repoProgress.parallel.files) {
         error(tr("Internal error: End: unexpected thread %1"
                  " > total files %2\n")
                  .arg(thread).arg(repoProgress.parallel.files));
      } else {
         const ThreadId&  tid = repoProgress.parallel.threads[thread];
         repoProgress.parallel.retryMapping.remove(tid);

         pg->setItemActsThreadValue(thread, current);
         pg->endItemActThread(thread);

         repoProgress.parallel.meterMapping.remove(meterObj);
      }
   }
}

void YumIO::collectFailure (void*  meterObj)

{
#if defined(DEBUG_COLLECT)
   printf("collectFailure %p <----------------------------------------\n",
          meterObj);
#endif

   if (repoProgress.parallel.begun) {
      Progress::Threads  thread = repoProgress.parallel.meterMapping[meterObj];

      if (thread > repoProgress.parallel.files) {
         error(tr("Internal error: Failure: unexpected thread %1"
                  " > total files %2\n")
                  .arg(thread).arg(repoProgress.parallel.files));
      } else {
         pg->retryItemActThread(thread);

         repoProgress.parallel.meterMapping.remove(meterObj);
      }
   }
}

PyObject*  YumIO::collectProgress(PyObject*  self, PyObject*  args)
{ 
   long       current;
   long       total;
   PyObject*  nameObj;
   
   if (!PyArg_ParseTuple(args, "llO", &current, &total, &nameObj)) return NULL;

   QString  name;
   if (PyString_Check(nameObj) || PyUnicode_Check(nameObj)) {
      name = pyString(nameObj);
   } else if (nameObj == Py_None) {
      name = "";
   } else {
      error(tr("Internal error: unexpected progress name"));
      name = "";
   }

   if (repoProgress.act > 0
       && total == repoProgress.total
       && name == repoProgress.name) {
#if defined(DEBUG_COLLECT_GRAB_METER_DETAIL)
      printf("collect: update: %lu/%lu\n", current, total);
#endif
      collectUpdate(NULL, current);
   } else if (current == -1 && total == -1) {
      // We were called by the YumIOSqlitePackageSack.addDict, which
      // indicates the end of a read of the sqlite database.
#if defined(DEBUG_COLLECT_GRAB_METER)
      printf("collect: start: %s RA_Read (sqlite database)\n",
             name.toStdString().c_str());
#endif
      collectStart(NULL, name, RA_Read, 1, 1);
   } else if (name == COLL_GRAB_METADATA) {
      if (repoProgress.act > 0
          && total == repoProgress.total
          && repoProgress.name.isEmpty()) {
#if defined(DEBUG_COLLECT_GRAB_METER_DETAIL)
         printf("collect: update: %lu/%lu %s (metadata)\n",
                current, total, name.toStdString().c_str());
#endif
         collectUpdate(NULL, current);
      } else {
         // mdcache: special name indicates a convert
         // For the cases where name is empty, I could use:
         //    repoProgress.repos[(repoProgress.act-1) / RA_Count]
         // However, that isn't 100% reliable because we don't get callbacks
         // for repositories that are completely empty.  So I don't do it.
#if defined(DEBUG_COLLECT_GRAB_METER)
         printf("collect: start: %lu/%lu %s RA_Convert (metadata: mdcache)\n",
                current, total, name.toStdString().c_str());
#endif
         collectStart(NULL, QString(), RA_Convert, current, total);
      }
   } else {
      if (usesSqlite()) {
         // sqlite: a straight name indicates a Convert
#if defined(DEBUG_COLLECT_GRAB_METER)
         printf("collect: start: %lu/%lu %s RA_Convert\n",
                current, total, name.toStdString().c_str());
#endif
         collectStart(NULL, name, RA_Convert, current, total);
      } else {
         // mdcache: a straight name indicates a Read
#if defined(DEBUG_COLLECT_GRAB_METER)
         printf("collect: start: %lu/%lu %s RA_Read\n",
                current, total, name.toStdString().c_str());
#endif
         collectStart(NULL, name, RA_Read, current, total);
      }
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::collectGrabMeterStart (PyObject*  self, PyObject*  args)
{
   PyObject*      meterObj;
   PyObject*      filenameObj;
   PyObject*      urlObj;
   PyObject*      simplenameObj;
   PyObject*      sizeObj;
   PyObject*      textObj;

   if (!PyArg_ParseTuple(args, "OOOOOO", 
                         &meterObj, &filenameObj, &urlObj, &simplenameObj,
                         &sizeObj, &textObj)) return NULL;

#if defined(DEBUG_COLLECT_GRAB_METER_ARGS)
   printf("collect: start: meter=%p %s filename=%s url=%s text=%s\n",
          meterObj,
          pyDescribe(sizeObj),
          pyDescribe(filenameObj),
          pyDescribe(urlObj),
          pyDescribe(textObj));
#endif

   Package::Size  size;
   if (PyLong_Check(sizeObj)) {
      size = PyLong_AsLongLong(sizeObj);
   } else if (PyInt_Check(sizeObj)) {
      size = PyInt_AsLong(sizeObj);
   } else if (sizeObj == Py_None) {
      size = 1;
   } else {
      warning(tr("Internal error: collect grab size not understood"));
   }

   QString  name;
   if (PyString_Check(filenameObj) || PyUnicode_Check(filenameObj)) {
      name = pyString(filenameObj);
   } else if (filenameObj == Py_None) {
      name = "";
   } else {
      error(tr("Internal error: unexpected progress filename"));
      name = "";
   }

   bool  use_text    = false; // name from text, as opposed to filename.

   bool  is_repomd   = false;
   bool  is_metalink = false;
   bool  is_group    = false;

   if (name.isEmpty()) {
      if (PyString_Check(textObj) || PyUnicode_Check(textObj)) {
         use_text = true;
         name = pyString(textObj);
      }
   }

   int  slash = name.lastIndexOf('/');

   if (slash != -1) {
      is_repomd   = (name.mid(slash+1, 6) == "repomd");
      is_metalink = (name.mid(slash+1, 8) == "metalink");
      is_group    = (name.mid(slash+1, 8) == "group_gz");
   }

   if (use_text) {
      // These often are of the form ${name}/${release}/${arch}
      int  slash = name.indexOf('/');
      if (slash != -1) {
         name.truncate(slash);
      }
   } else {
      int  slash2 = name.lastIndexOf('/', slash-1);
      if (slash != -1 && slash2 != -1) {
         name.truncate(slash);
         name.remove(0, slash2+1);
      }
      name.remove(QRegExp("buffet\?.*=.*&.*=.*&path="));
   }
   
#if defined(DEBUG_COLLECT_GRAB_METER)
   printf("collect: start: %llu %s\n", size, name.toStdString().c_str());
#endif

   collectStart(meterObj,
                name, 
                (is_metalink ? RA_Grab_Metalink :
                 is_repomd   ? RA_Grab_RepoMD   : 
                 is_group    ? RA_Grab_Group    :
                               RA_Grab),
                0, 
                size);

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::collectGrabMeterUpdate (PyObject*  self, PyObject*  args)
{
   PyObject*  meterObj;
   long long  amount;

   if (!PyArg_ParseTuple(args, "OL", &meterObj, &amount)) return NULL;

#if defined(DEBUG_COLLECT_GRAB_METER_DETAIL)
   printf("collect: update: meter=%p %llu\n", meterObj, amount);
#endif

   collectUpdate(meterObj, amount);

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::collectGrabMeterEnd (PyObject*  self, PyObject*  args)
{
   PyObject*  meterObj;
   long long  amount;

   if (!PyArg_ParseTuple(args, "OL", &meterObj, &amount)) return NULL;

#if defined(DEBUG_COLLECT_GRAB_METER)
   printf("collect: %p %llu end\n", meterObj, amount);
#endif

   collectEnd(meterObj, amount);

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::collectGrabMeterFailure (PyObject*  self, PyObject*  args)
{
   PyObject*  meterObj;
   PyObject*  messageObj;

   if (!PyArg_ParseTuple(args, "OO", &meterObj, &messageObj)) return NULL;

   QString  message;
   if (PyString_Check(messageObj) || PyUnicode_Check(messageObj)) {
      message = pyString(messageObj);
   } else if (messageObj == Py_None) {
      message = "";
   } else {
      error(tr("Internal error: unexpected progress failure message"));
      message = "";
   }

#if defined(DEBUG_COLLECT_GRAB_METER)
   printf("collect: %p failure %s\n", meterObj, message.toStdString().c_str());
#endif

   collectFailure(meterObj);

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::collectGrabMultiMeterStart (PyObject*  self, PyObject*  args)
{
   long long  total_files;
   long long  total_size;

   if (!PyArg_ParseTuple(args, "LL", &total_files, &total_size)) return NULL;

#if defined(DEBUG_COLLECT_GRAB_METER_MULTI)
   printf("collect: Multi start: total_files=%llu total_size=%llu\n",
          total_files, total_size);
#endif

   // We don't know which activities are involved in this parallel block
   // of acts until we see the first collectGrabMeterStart.  So, just flag
   // that it has begun, and let collectStart() do the pg->setItemActsThread
   // when it sees the first activity.
   repoProgress.parallel.begun          = true;
   repoProgress.parallel.setup          = false;
   repoProgress.parallel.files          = total_files;
   repoProgress.parallel.size           = total_size;
   repoProgress.parallel.threadsStarted = 0;
   repoProgress.parallel.meterMapping.clear();
   repoProgress.parallel.threads        = new ThreadId[total_files];
   repoProgress.parallel.retryMapping.clear();

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::collectGrabMultiMeterNew (PyObject*  self, PyObject*  args)
{
   PyObject*  meterObj;

   if (!PyArg_ParseTuple(args, "O", &meterObj)) return NULL;

#if defined(DEBUG_COLLECT_GRAB_METER_MULTI)
   printf("collect: Multi new: meter=%p\n", meterObj);
#endif

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::collectGrabMultiMeterRemove (PyObject*  self, PyObject*  args)
{
   PyObject*  meterObj;

   if (!PyArg_ParseTuple(args, "O", &meterObj)) return NULL;

#if defined(DEBUG_COLLECT_GRAB_METER_MULTI)
   printf("collect: Multi remove: meter=%p\n", meterObj);
#endif

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::collectGrabMultiMeterEnd (PyObject*  self, PyObject*  args)
{
#if defined(DEBUG_COLLECT_GRAB_METER_MULTI)
   printf("collect: Multi end\n");
#endif

   pg->endItemActsThreads();

   repoProgress.parallel.begun          = false;
   repoProgress.parallel.setup          = false;
   repoProgress.parallel.threadsStarted = 0;
   repoProgress.parallel.meterMapping.clear();
   delete [] repoProgress.parallel.threads;
   repoProgress.parallel.threads        = NULL;
   repoProgress.parallel.retryMapping.clear();

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::applyGrabMeterStart (PyObject*  self, PyObject*  args)
{
   PyObject*  meterObj;
   PyObject*  filenameObj;
   PyObject*  urlObj;
   PyObject*  simplenameObj;
   PyObject*  sizeObj;
   PyObject*  textObj;

   if (!PyArg_ParseTuple(args, "OOOOOO", 
                         &meterObj, &filenameObj, &urlObj, &simplenameObj,
                         &sizeObj, &textObj)) return NULL;

#if defined(DEBUG_APPLY_GRAB_METER_ARGS)
   printf("apply: start: meter=%p %s filename=%s text=%s\n",
          meterObj,
          pyDescribe(sizeObj),
          pyDescribe(filenameObj),
          pyDescribe(textObj));
#endif

   Package::Size  size;
   if (PyLong_Check(sizeObj)) {
      size = PyLong_AsLongLong(sizeObj);
   } else if (PyInt_Check(sizeObj)) {
      size = PyInt_AsLong(sizeObj);
   } else if (sizeObj == Py_None) {
      size = 1;
   } else {
      warning(tr("Internal error: collect grab size not understood"));
   }

   QString  text;
   if (PyString_Check(textObj) || PyUnicode_Check(textObj)) {
      text = pyString(textObj);
   } else if (textObj == Py_None) {
      text = "";
   } else {
      error(tr("Internal error: unexpected progress text type"));
      text = "";
   }

   if (text == "<delta rebuild>") {
#if defined(DEBUG_APPLY_GRAB_METER)
      printf("apply: start: %s size=%llu\n", 
             text.toStdString().c_str(), size);
#endif
      pg->endItemActs(); // PG_R_GRAB_RPMS
      ts.run.deltaRebuilt = 1;
      pg->startItem(tr(PG_R_DELTA_REBUILD_TEXT), PG_R_DELTA_REBUILD, size);
   } else {
      QString  filename;
      if (PyString_Check(filenameObj) || PyUnicode_Check(filenameObj)) {
         filename = pyString(filenameObj);
      } else if (filenameObj == Py_None) {
         filename = "";
      } else {
         error(tr("Internal error: unexpected progress name type"));
         filename = "";
      }

      QString  name = filename;
      if (name.isEmpty()) {
         name = text;
      }

      int  slash = name.lastIndexOf('/');
      if (slash != -1) {
         name.remove(0, slash+1);
      }
      name.remove(QRegExp("buffet\?.*=.*&.*=.*&path="));
      if (name.endsWith(".rpm") || name.endsWith(".hdr")) {
         name.chop(4);
      }

#if defined(DEBUG_APPLY_GRAB_METER)
      printf("apply: start: %llu %s\n", size, name.toStdString().c_str());
#endif

      QString  activity;
      if (ts.confirmed) {
         activity = tr("Downloading %1").arg(name);
      } else {
         activity = tr("Downloading %1 header").arg(name);
      }
      if (ts.parallel.begun) {
         Progress::Threads  thread = ts.parallel.meterMapping[meterObj];
         pg->startItemActThread(activity, thread, size);
      } else {
         if (ts.confirmed) {
            // All rpms are not the same size, so use size as the number of acts
            ts.act += PG_R_GRAB(size);
         } else {
            // Assume that all headers are roughly the same size and use number
            // of grabs as the number of acts.
            ts.act += PG_D_GRAB(1);
         }
         pg->startItemAct(activity, size, ts.act);
      }
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::applyGrabMeterUpdate (PyObject*  self, PyObject*  args)
{
   PyObject*  meterObj;
   long long  amount;

   if (!PyArg_ParseTuple(args, "OL", &meterObj, &amount)) return NULL;

#if defined(DEBUG_APPLY_GRAB_METER_DETAIL)
   printf("apply: update: meter=%p %llu\n", meterObj, amount);
#endif

   if (ts.parallel.begun) {
      Progress::Threads  thread = ts.parallel.meterMapping[meterObj];
      pg->setItemActsThreadValue(thread, amount);
   } else {
      pg->setItemValue(amount);
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::applyGrabMeterEnd (PyObject*  self, PyObject*  args)
{
   PyObject*  meterObj;
   long long  amount;

   if (!PyArg_ParseTuple(args, "OL", &meterObj, &amount)) return NULL;

#if defined(DEBUG_APPLY_GRAB_METER)
   printf("apply: %p %llu end\n", meterObj, amount);
#endif

   if (ts.parallel.begun) {
      Progress::Threads  thread = ts.parallel.meterMapping[meterObj];
      pg->setItemActsThreadValue(thread, amount);
      pg->endItemActThread(thread);
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::applyGrabMultiMeterStart (PyObject*  self, PyObject*  args)
{
   long long      total_files;
   Package::Size  total_size;

   if (!PyArg_ParseTuple(args, "LL", &total_files, &total_size)) return NULL;

#if defined(DEBUG_APPLY_GRAB_METER_MULTI)
   printf("apply: Multi start: total_files=%llu total_size=%llu\n",
          total_files, total_size);
#endif

   // We don't know which activities are involved in this parallel block
   // of acts until we see the first applyGrabMeterStart.  So, just flag
   // that it has begun, and let applyGrabMeterStart() do the 
   // pg->setItemActsThread when it sees the first activity.
   ts.parallel.begun          = true;
   ts.parallel.files          = total_files;
   ts.parallel.size           = total_size;
   ts.parallel.threadsStarted = 0;
   ts.parallel.meterMapping.clear();

   QString  activity = QString("Downloading %1 packages").arg(ts.parallel.files);
   Progress::Steps  firstAct = ts.act;
   if (ts.confirmed) {
      ts.act += PG_R_GRAB(total_size);
   } else {
      ts.act += PG_D_GRAB(1);
   }
   Progress::Steps  tailAct = ts.act;
   pg->setupItemActsThreads(activity,
                            firstAct,
                            tailAct,
                            ts.parallel.files,
                            ts.parallel.size);

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::applyGrabMultiMeterNew (PyObject*  self, PyObject*  args)
{
   PyObject*  meterObj;

   if (!PyArg_ParseTuple(args, "O", &meterObj)) return NULL;

#if defined(DEBUG_APPLY_GRAB_METER_MULTI)
   printf("apply: Multi new: meter=%p\n", meterObj);
#endif

   Progress::Threads  thread = ts.parallel.threadsStarted++;
   ts.parallel.meterMapping[meterObj] = thread;

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::applyGrabMultiMeterRemove (PyObject*  self, PyObject*  args)
{
   PyObject*  meterObj;

   if (!PyArg_ParseTuple(args, "O", &meterObj)) return NULL;

#if defined(DEBUG_APPLY_GRAB_METER_MULTI)
   printf("apply: Multi remove: meter=%p\n", meterObj);
#endif

   ts.parallel.meterMapping.remove(meterObj);

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::applyGrabMultiMeterEnd (PyObject*  self, PyObject*  args)
{
#if defined(DEBUG_APPLY_GRAB_METER_MULTI)
   printf("apply: Multi end\n");
#endif

   pg->endItemActsThreads();

   ts.parallel.begun          = false;
   ts.parallel.threadsStarted = 0;
   ts.parallel.meterMapping.clear();

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::dsTsCheck(PyObject*  self, PyObject*  args)
{ 
#if defined(DEBUG_TS)
   printf("dsTsCheck\n");
#endif

   ts.act += ts.dep.depSolveActs;
   pg->startItemAct(tr("Resolve Dependencies"), 1, ts.act);

   Py_INCREF(Py_None);
   return Py_None;
}

void  YumIO::dsTsInfo (char*  msg)
{
#if defined(DEBUG_TS)
   printf("dsTsInfo\n");
#endif

   if (strncmp(msg, 
               APPLY_DEPSOLVE_UPDATING, 
               STRLEN(APPLY_DEPSOLVE_UPDATING)) == 0) {
      QString  name(msg + STRLEN(APPLY_DEPSOLVE_UPDATING));
      int      space = name.indexOf(' ');
      if (space == -1) {
         error(tr("Internal error: unexpected TSINFO: %1").arg(msg));
         return;
      }
      name.resize(space);
      Package::Entry*  entry = dsTsFindPkg(name);
      dsTsNewPkg(name, entry, 2, true);
   } else if (strncmp(msg, 
                      APPLY_DEPSOLVE_OBSOLETING, 
                      STRLEN(APPLY_DEPSOLVE_OBSOLETING)) == 0) {
      QString  oldName(msg + STRLEN(APPLY_DEPSOLVE_OBSOLETING));
      int      oldSpace = oldName.indexOf(' ');
      if (oldSpace == -1) {
         error(tr("Internal error: unexpected TSINFO: %1").arg(msg));
         return;
      }
      int      withIx = oldName.indexOf(APPLY_DEPSOLVE_OBSOLETING_WITH, 
                                        oldSpace + 1);
      if (withIx == -1) {
         error(tr("Internal error: unexpected TSINFO: %1").arg(msg));
         return;
      }
      QString  newName(oldName
                       + withIx + STRLEN(APPLY_DEPSOLVE_OBSOLETING_WITH));
      int      newSpace = newName.indexOf(' ');
      if (newSpace == -1) {
         error(tr("Internal error: unexpected TSINFO: %1").arg(msg));
         return;
      }
      oldName.resize(oldSpace);
      newName.resize(newSpace);
      Package::Entry*  oldEntry = dsTsFindPkg(oldName);
      Package::Entry*  newEntry = dsTsFindPkg(newName);
      dsTsNewPkg(oldName, oldEntry, 1, false);
      dsTsNewPkg(newName, newEntry, 1, true);
   } else if (strncmp(msg, 
                      APPLY_DEPSOLVE_MARKING, 
                      STRLEN(APPLY_DEPSOLVE_MARKING)) == 0) {
      QString  name(msg + STRLEN(APPLY_DEPSOLVE_MARKING));
      int      space = name.indexOf(' ');
      if (space == -1) {
         error(tr("Internal error: unexpected TSINFO: %1").arg(msg));
         return;
      }
      name.resize(space);
      bool  isUpdate = strstr(msg, APPLY_DEPSOLVE_MARKING_UPDATE) != NULL;
      Package::Entry*  entry = dsTsFindPkg(name);
      dsTsNewPkg(name, entry, isUpdate ? 2 : 1, true);
   } else if (strstr(msg, APPLY_DEPSOLVE_ERASE) != NULL) {
      // DO NOTHING
   } else {
      error(tr("Internal error: unexpected TSINFO: %1").arg(msg));
      return;
   }
}

Package::Entry*  YumIO::dsTsFindPkg (QString  name)
{
   Package::Entry*  entry;
   if (yumVersion() >= 30209) {
      // The name actually is full a envra ([epoch:]name-version-release.arch).
      // Parse it to get an entry for a specific architecture.  (I don't need
      // the epoch:version-release for this, so I ignore it.)
      QString  simpleName;
      QString  arch;
      if (parseEnvra(name, simpleName, arch)) {
         entry = hash[EntryKey(false, simpleName, getBiArch(arch))];
      } else {
         error(tr("Internal error:"
                  " TSINFO mentions package %1 which does not appear to be a"
                  " valid envra; interpreting as package name").arg(name));
         entry = findPackage(name);
      }
   } else {
      entry = findPackage(name);
   }

   if (entry == NULL) {
      error(tr("Internal error:"
               " TSINFO mentions unknown package: %1").arg(name));
   }

   return entry;
}

void  YumIO::dsTsNewPkg (QString          name,
                         Package::Entry*  entry,
                         unsigned int     numTs,
                         bool             needHdr)
{
#if defined(DEBUG_TS)
   printf("dsTsNewPkg\n");
#endif

   if (entry == NULL) return;

   if (entry->state == Package::INSTALLED) {
      error(tr("Internal error:"
               " TSINFO mentions intent to install or upgrade package that"
               " already seems up-to-date: %1").arg(entry->name));
      return;
   }

   ts.dep.totalTs += numTs;
   if (needHdr && !QFile::exists(getLocalHdr(entry->obj))) {
      ts.dep.totalGrabs++;
   }
   unsigned int  grabActs = PG_D_GRAB(ts.dep.totalGrabs);
   unsigned int  numActs  = grabActs + ts.dep.totalDepSolveActs;
   pg->changeItemActs(numActs); // PG_D_DEPSOLVE
}


PyObject*  YumIO::dsRestartLoop(PyObject*  self, PyObject*  args)
{ 
#if defined(DEBUG_TS)
   printf("dsRestartLoop\n");
#endif

   // We'll be doing another depsolve, so re-estimate and add the new
   // estimate.
   unsigned int  grabActs = PG_D_GRAB(ts.dep.totalGrabs);
   ts.dep.depSolveActs       = PG_D_DEPSOLVE_ESTIMATE(ts.dep.totalTs);
   ts.dep.totalDepSolveActs += ts.dep.depSolveActs;
   unsigned int  numActs = grabActs + ts.dep.totalDepSolveActs;
   pg->changeItemActs(numActs); // PG_D_DEPSOLVE

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::userconfirm(PyObject*  self, PyObject*  args)
{ 
#if defined(DEBUG_TS)
   printf("userconfirm\n");
#endif

   pg->endItemActs(); // PG_D_DEPSOLVE

   pg->reset(); // PG_D_*

   int  result = dialogs->confirm(tr("Apply"), ts.descr);

   ts.confirmed = (result == QDialog::Accepted);

   if (ts.confirmed) {
      log(Log::ACTION, 0, tr("Applying: OK"));
   } else {
      log(Log::ACTION, 0, tr("Not Applying: Canceled"));
   }

   if (ts.confirmed) {
      // The PG_R_GRAB_RPMS_NOMINAL fractional part of PG_R_TOTAL is based
      // on actually having to grab all the rpm's.  If some already are in
      // the cache, then scale it back accordinly.
      ts.run.pgRGrabRpms = (ts.run.nominalGrabSize == 0
                            ? 0
                            : (ts.run.totalGrabSize * PG_R_GRAB_RPMS_NOMINAL
                               / ts.run.nominalGrabSize));

      pg->startTotal(tr(PG_R_TOTAL_TEXT), PG_R_TOTAL);

      pg->startItem(tr(PG_R_INITIALIZE_TEXT), PG_R_INITIALIZE);

      // All rpms are not the same size, so use size as the number of acts
      pg->setupItemActs(ts.run.pgRGrabRpms, ts.run.totalGrabSize);
   }

   return PyInt_FromLong(ts.confirmed);
}

PyObject*  YumIO::reportDownloadSize(PyObject*  self, PyObject*  args)
{ 
   PyObject*  packages;
   
   if (!PyArg_ParseTuple(args, "O", &packages)) return NULL;

#if defined(DEBUG_TS)
   printf("reportDownloadSize\n");
#endif

   Package::Size  size = 0;

   PyObject*  it = PyObject_GetIter(packages);
   for (;;) {
      PyObject*  pkg = PyIter_Next(it);
      if (pkg == NULL) break;
      size += sizeOf(pkg);
      Py_DECREF(pkg);
   }
   
   Py_DECREF(it);

   ts.descr.append(tr("Total download size: %1\n").arg(bytesFormat(size)));

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::listTransaction(PyObject*  self, PyObject*  args)
{ 
   PyObject*  yumIO;
   
   if (!PyArg_ParseTuple(args, "O", &yumIO)) return NULL;

#if defined(DEBUG_TS)
   printf("listTransaction\n");
#endif

   PyObject*  tsInfo;
   PyObject*  lists;
   PyObject*  updated;
   PyObject*  installed;
   PyObject*  removed;
   PyObject*  obsoleted;
   PyObject*  depup;
   PyObject*  depinst;
   PyObject*  deprem;

   if (!makeLists(yumIO, tsInfo, lists, 
                  updated, installed, removed, obsoleted, 
                  depup, depinst, deprem)) {
      return NULL;
   }

   ts.descr.append(tr("Transaction to be performed:\n"));

   if (   !displayTransactionPkgsRepo(tr("Install"), installed)
       || !displayTransactionPkgsRepo(tr("Update"),  updated)
       || !displayTransactionPkgs    (tr("Remove"),  removed)) {
      Py_DECREF(lists);
      Py_DECREF(tsInfo);
      return NULL;
   }

   if (   PySequence_Length(depinst) > 0
       || PySequence_Length(depup) > 0
       || PySequence_Length(deprem) > 0) {
      ts.descr.append(tr("Performing the following to resolve"
                         " dependencies:\n"));
      if (   !displayTransactionPkgsRepo(tr("Install"), depinst)
          || !displayTransactionPkgsRepo(tr("Update"),  depup)
          || !displayTransactionPkgs    (tr("Remove"),  deprem)) {
         Py_DECREF(lists);
         Py_DECREF(tsInfo);
         return NULL;
      }
   }

   if (PySequence_Length(obsoleted) > 0) {
      ts.descr.append(tr("Other Transactions:\n"));
      if (!displayTransactionPkgsObs(tr("Obsoleting"), obsoleted)) {
         Py_DECREF(lists);
         Py_DECREF(tsInfo);
         return NULL;
      }
   }

   noteTransactionPkgs(Package::INSTALL, installed);
   noteTransactionPkgs(Package::UPDATE,  updated);
   noteTransactionPkgs(Package::REMOVE,  removed);

   noteTransactionPkgs(Package::INSTALL, depinst);
   noteTransactionPkgs(Package::UPDATE,  depup);
   noteTransactionPkgs(Package::REMOVE,  deprem);

   noteTransactionPkgs(Package::REMOVE,  obsoleted);

   Py_DECREF(lists);
   Py_DECREF(tsInfo);

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::tsInitCallback(PyObject*  self, PyObject*  args)
{
   long  output;

   if (!PyArg_ParseTuple(args, "l", &output)) return NULL;

#if defined(DEBUG_TS)
   printf("tsInitCallback (output=%lu)\n", output);
#endif

   ts.act = 0;

   if (output) {
      if (hasTransTest()) {
         pg->endItemActs(); // for PG_R_TRANS_TEST
      }

      pg->startItem(tr(PG_R_RUN_TRANS_INIT_TEXT), PG_R_RUN_TRANS_INIT);
      // yum does another transaction test as part of the transaction
      // (or in 3.2.3 or later, this is the only transaction test)
      Progress::Steps  testActs = PG_R_TEST_TRANS_ESTIMATE(ts.run.totalTsSize);
      Progress::Steps  numActs  = ts.run.totalTsSize + testActs;
      pg->setupItemActs(PG_R_RUN_TRANS, numActs);
   } else {
      if (!ts.run.deltaRebuilt) {
         pg->endItemActs(); // for PG_R_GRAB_RPMS
         if (usesPresto()) {
            pg->startItem(tr(PG_R_DELTA_REBUILD_TEXT), PG_R_DELTA_REBUILD);
            pg->setItemValue(1);
         }
      }

      if (hasTransTest()) {
         pg->startItem(tr(PG_R_TRANS_TEST_INIT_TEXT), PG_R_TRANS_TEST_INIT);
         Progress::Steps  numActs
            = PG_R_TEST_TRANS_ESTIMATE(ts.run.totalTsSize);
         pg->setupItemActs(PG_R_TRANS_TEST, numActs);
      }
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::tsCallback(PyObject*  self, PyObject*  args)
{
   // This routine is called in yum <= 3.2.2.  tsEvent is called instead in
   // yum >= 3.2.3.

   long             what;
   Progress::Steps  progress;
   Progress::Steps  total;
   PyObject*        h;
   PyObject*        user;
   bool             understood = false;

   if (!PyArg_ParseTuple(args, "lLLOO", 
                         &what, &progress, &total, &h, &user)) return NULL;

#if defined(DEBUG_TS_DETAIL)
   QString  whatStr;
   if (what & RPMCALLBACK_INST_OPEN_FILE)  whatStr.append(" INST_OPEN_FILE");
   if (what & RPMCALLBACK_INST_CLOSE_FILE) whatStr.append(" INST_CLOSE_FILE");
   if (what & RPMCALLBACK_TRANS_START)     whatStr.append(" TRANS_START");
   if (what & RPMCALLBACK_TRANS_PROGRESS)  whatStr.append(" TRANS_PROGRESS");
   if (what & RPMCALLBACK_TRANS_STOP)      whatStr.append(" TRANS_STOP");
   if (what & RPMCALLBACK_INST_START)      whatStr.append(" INST_START");
   if (what & RPMCALLBACK_INST_PROGRESS)   whatStr.append(" INST_PROGRESS");
   if (what & RPMCALLBACK_UNINST_START)    whatStr.append(" UNINST_START");
   if (what & RPMCALLBACK_UNINST_PROGRESS) whatStr.append(" UNINST_PROGRESS");
   if (what & RPMCALLBACK_UNINST_STOP)     whatStr.append(" UNINST_STOP");
   whatStr = whatStr.trimmed();
   printf("tsCallback (what=%s, progress=%llu/%llu)\n", 
          whatStr.toStdString().c_str(), progress, total);
#endif

   if (what & RPMCALLBACK_INST_OPEN_FILE) {
      // DO NOTHING: handled by Python code
      understood = true;
   }
   if (what & RPMCALLBACK_INST_CLOSE_FILE) {
      // DO NOTHING: handled by Python code
      understood = true;
   }
   if (what & RPMCALLBACK_TRANS_START) {
      ts.act += PG_R_TEST_TRANS_ESTIMATE(ts.run.totalTsSize);
      pg->startItemAct(tr("Transaction Test"), total, ts.act);
      understood = true;
   }
   if (what & RPMCALLBACK_TRANS_PROGRESS) {
      pg->setItemValue(progress);
      understood = true;
   }
   if (what & RPMCALLBACK_TRANS_STOP) {
      pg->setItemValue(total);
      understood = true;
   }
   if (what & RPMCALLBACK_INST_START) {
      PyObject*  hdr   = PyTuple_GetItem(h, 0);
      QString    name  = pyMappingString(hdr, "name");
      QString    arch  = pyMappingString(hdr, "arch");
#if defined(DEBUG_TS_DETAIL)
      printf("           (name=%s) (arch=%s)\n",
             name.toStdString().c_str(),
             arch.toStdString().c_str());
#endif
      Package::Entry*  entry = hash[EntryKey(false, name, getBiArch(arch))];
      QString          activity;
      if (entry->apply.action == Package::UPDATE) {
         activity = tr("Updating %1").arg(name);
         ts.act += PG_R_UPDATE(entry->size);
      } else {
         activity = tr("Installing %1").arg(name);
         ts.act += PG_R_INSTALL(entry->size);
      }
      // RPMInstallCallback.callback fails to say anything for this case,
      // so do it manually here.
      log(Log::NORMAL, 1, activity);
      pg->startItemAct(activity, total, ts.act);
      understood = true;
   }
   if (what & RPMCALLBACK_INST_PROGRESS) {
      pg->setItemValue(progress);
      understood = true;
   }
   if (what & RPMCALLBACK_UNINST_START) {
      QString  name  = pyString(h);
#if defined(DEBUG_TS_DETAIL)
      printf("           (name=%s))\n", name.toStdString().c_str());
#endif
      Package::Entry*  entry = findPackageAction(name, Package::UPDATE);
      bool             isCompletingUpdate;
      if (entry != NULL) {
         isCompletingUpdate = true;
      } else {
         entry = findPackageAction(name, Package::REMOVE);
         if (entry == NULL) {
            warning(tr("Internal error: YumIO::tsCallback UNINST of %1"
                       " not expected").arg(name));
            entry = findPackage(name);
         }
         isCompletingUpdate = false;
      }
      QString  activity;
      if (isCompletingUpdate) {
         // update: "removal" of old package
         activity = tr("Completing update for: %1").arg(name);
         ts.act += PG_R_COMPLETE_UPDATE(entry->next->size);
      } else {
         activity = tr("Removing %1").arg(name);
         ts.act += PG_R_REMOVE(entry->size);
      }
      pg->startItemAct(activity, total, ts.act);
      understood = true;
   }
   if (what & RPMCALLBACK_UNINST_PROGRESS) {
      pg->setItemValue(progress);
      understood = true;
   }
   if (what & RPMCALLBACK_UNINST_STOP) {
      pg->setItemValue(total);
      understood = true;
   }

   if (!understood) {
      warning(tr("Internal error: YumIO::tsCallback what=%1 not understood;"
                 " ignoring").arg(what));
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::tsEvent(PyObject*  self, PyObject*  args)
{
   // This routine is called in yum >= 3.2.3.  tsCallback is called instead in
   // yum <= 3.2.2.

   PyObject*        pkg;
   long             action;
   Progress::Steps  progress;
   Progress::Steps  total;
   unsigned long    currentTs;
   unsigned long    totalTs;
   bool             understood = false;

   if (!PyArg_ParseTuple(args, "OlLLll",
                         &pkg, &action,
                         &progress, &total,
                         &currentTs, &totalTs)) return NULL;

#if defined(DEBUG_TS_DETAIL)
   QString  actStr;
   if      (action == TS_UPDATE)      actStr = "TS_UPDATE";
   else if (action == TS_INSTALL)     actStr = "TS_INSTALL";
   else if (action == TS_TRUEINSTALL) actStr = "TS_TRUEINSTALL";
   else if (action == TS_ERASE)       actStr = "TS_ERASE";
   else if (action == TS_OBSOLETED)   actStr = "TS_OBSOLETED";
   else if (action == TS_OBSOLETING)  actStr = "TS_OBSOLETING";
   else if (action == TS_AVAILABLE)   actStr = "TS_AVAILABLE";
   else if (action == TS_UPDATED)     actStr = "TS_UPDATED";
   actStr = actStr.trimmed();
   printf("tsEvent (action=%s, progress=%llu/%llu)\n", 
          actStr.toStdString().c_str(), progress, total);
#endif

   if (IS_INSTALL(action) || IS_UPDATE(action)) {
      if (currentTs != ts.run.lastTs) {
         QString  name  = pyStringDec(PyObject_GetAttrString(pkg, "name"));
         QString  arch  = pyStringDec(PyObject_GetAttrString(pkg, "arch"));
#if defined(DEBUG_TS_DETAIL)
         printf("           (name=%s) (arch=%s)\n",
                name.toStdString().c_str(),
                arch.toStdString().c_str());
#endif
         Package::Entry*  entry = hash[EntryKey(false, name, getBiArch(arch))];
         QString          activity;
         if (IS_UPDATE(action)) {
            activity = tr("Updating %1").arg(name);
            ts.act += PG_R_UPDATE(entry->size);
         } else {
            activity = tr("Installing %1").arg(name);
            ts.act += PG_R_INSTALL(entry->size);
         }
         log(Log::NORMAL, 1, activity);
         pg->startItemAct(activity, total, ts.act);
         ts.run.lastTs = currentTs;
      }
      pg->setItemValue(progress);
      understood = true;
   } else if (IS_REMOVE(action) || IS_COMPLETING_UPDATE(action)) {
      if (currentTs != ts.run.lastTs) {
         QString  name  = pyString(pkg);
#if defined(DEBUG_TS_DETAIL)
         printf("        (name=%s)\n", name.toStdString().c_str());
#endif
         bool             isCompletingUpdate = IS_COMPLETING_UPDATE(action);
         Package::Entry*  entry              = NULL;
         if (isCompletingUpdate) {
            entry = findPackageAction(name, Package::UPDATE);
            if (entry == NULL) {
               // This could be a case of installing pkgname.arch1 while
               // simultaneously removing pkgname.arch2.  This confuses yum into
               // claiming the removal of pkgname.arch2 is an update completion
               // for the vague pkgname (with no arch).  Fix it to avoid a 
               // potential SIGSEGV (because pkgname.arch2 probably has no 
               // entry->next).
               isCompletingUpdate = false;
            }
         }
         if (entry == NULL) {
            entry = findPackageAction(name, Package::REMOVE);
         }
         if (entry == NULL) {
            warning(tr("Internal error: YumIO::tsCallback REMOVE/COMPLETE of %1"
                       " not expected").arg(name));
            entry = findPackage(name);
         }
         QString  activity;
         if (isCompletingUpdate) {
            activity = tr("Completing update for: %1").arg(name);
            ts.act += PG_R_COMPLETE_UPDATE(entry->next->size);
         } else {
            activity = tr("Removing %1").arg(name);
            ts.act += PG_R_REMOVE(entry->size);
         }
         pg->startItemAct(activity, total, ts.act);
         ts.run.lastTs = currentTs;
      }
      pg->setItemValue(progress);
      understood = true;
   } else if (action == TS_AVAILABLE) {
      // DO NOTHING
      understood = true;
   }

   if (!understood) {
      warning(tr("Internal error: YumIO::tsEvent action=%1 not understood;"
                 " ignoring").arg(action));
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::postTransactionOutput(PyObject*  self, PyObject*  args)
{ 
   PyObject*  yumIO;
   
   if (!PyArg_ParseTuple(args, "O", &yumIO)) return NULL;

   pg->endItemActs(); // for PG_R_RUN_TRANS

   PyObject*  tsInfo;
   PyObject*  lists;
   PyObject*  updated;
   PyObject*  installed;
   PyObject*  removed;
   PyObject*  obsoleted;
   PyObject*  depup;
   PyObject*  depinst;
   PyObject*  deprem;

   if (!makeLists(yumIO, tsInfo, lists, 
                  updated, installed, removed, obsoleted, 
                  depup, depinst, deprem)) {
      return NULL;
   }

#if defined(DEBUG_TS)
   printf("postTransactionOutput\n");
#endif

   if (   !displayTransactionPkgs(tr("Removed"),              removed)
       || !displayTransactionPkgs(tr("Dependency Removed"),   deprem)
       || !displayTransactionPkgs(tr("Installed"),            installed)
       || !displayTransactionPkgs(tr("Dependency Installed"), depinst)
       || !displayTransactionPkgs(tr("Updated"),              updated)
       || !displayTransactionPkgs(tr("Dependency Updated"),   depup)
       || !displayTransactionPkgs(tr("Obsoleted"),            obsoleted)) {
      Py_DECREF(lists);
      Py_DECREF(tsInfo);
      return NULL;
   }

   Py_DECREF(lists);
   Py_DECREF(tsInfo);

   return PyString_FromString("");
}

/*
** Utility functions
*/

QString  YumIO::getLocalPkg (PyObject*  pkg)
{
   PyObject*  obj = pyMethodCall(pkg, "localPkg", "()");
   if (obj == NULL || obj == Py_None) return QString();
   return pyStringDec(obj);
}

QString  YumIO::getLocalHdr (PyObject*  pkg)
{
   PyObject*  obj = pyMethodCall(pkg, "localHdr", "()");
   if (obj == NULL || obj == Py_None) return QString();
   return pyStringDec(obj);
}

Package::Kind  YumIO::getPkgKind (PyObject*  pkg)
{
   return simpleString(pkg, "group") == "Suite" ? Package::SUITE 
                                                : Package::PACKAGE;
}

PyObject*  YumIO::getGroupItem_22 (const Package::Entry&  entry,
                                   char*                  attr)
{
   if (hasDistribGroups()) {
      error(tr("getGroupItem_22 called for incorrect version of yum"));
      return NULL;
   }

   PyObject*  dict = PyObject_GetAttrString(entry.obj, attr);
   if (dict == NULL) {
      error(tr("Internal error: groupInfo.%1 not evaluable").arg(attr));
      return NULL;
   }
   PyObject*  result = pyDictGetItem(dict, entry.name);
   if (result == NULL) {
      error(tr("Internal error: groupInfo.%1[%2]"
               " not evaluable").arg(attr).arg(entry.name));
      return NULL;
   }
   Py_INCREF(result); // return a new reference to be like getGroupItem_25
   Py_DECREF(dict);
   return result;
}

PyObject*  YumIO::getGroupItem_25 (const Package::Entry&  entry,
                                   char*                  attr)
{
   if (!hasDistribGroups()) {
      error(tr("getGroupItem_25 called for incorrect version of yum"));
      return NULL;
   }
      
   PyObject*  result = PyObject_GetAttrString(entry.obj, attr);
   if (result == NULL) {
      error(tr("Internal error: group %1.%2 not evaluable").arg(attr)
                                                           .arg(entry.name));
      return NULL;
   }
   return result;
}

Package::Entry*  YumIO::findPackage (QString  name)
{
   // Find a package given only the package name an no architecture information
   // (e.g. from a package group's list of packages).  First try the package for
   // the basearch.  If not found, pick any package with that name by checking
   // each architecture ever seen in the latest collect pass.

   EntryKey  key(false, name, getBiArch(basearch));
   if (hash.contains(key)) {
      return hash[key];
   }

   StringSet::const_iterator  it = biarches.begin();
   for (; it != biarches.end(); ++it) {
      if (*it != basearch) {
         EntryKey  key(false, name, *it);
         if (hash.contains(key)) {
            return hash[key];
         }
      }
   }

   return NULL;
}

Package::Entry*  YumIO::findPackageAction (QString          name,
                                           Package::Action  action)
{
   // Find a package given only the package name an no architecture information
   // (e.g. from a package group's list of packages).  First try the package for
   // the basearch.  If not found, pick any package with that name by checking
   // each architecture ever seen in the latest collect pass.

   EntryKey  key(false, name, getBiArch(basearch));
   if (hash.contains(key)) {
      Package::Entry*  entry = hash[key];
      if (entry->apply.action == action) {
         return entry;
      }
   }

   StringSet::const_iterator  it = biarches.begin();
   for (; it != biarches.end(); ++it) {
      if (*it != basearch) {
         EntryKey  key(false, name, *it);
         if (hash.contains(key)) {
            Package::Entry*  entry = hash[key];
            if (entry->apply.action == action) {
               return entry;
            }
         }
      }
   }

   return NULL;
}

bool  YumIO::parseEnvra (QString   envra,
                         QString&  name,
                         QString&  epoch,
                         QString&  version,
                         QString&  release,
                         QString&  arch)
{
   // The envra is in this format: [epoch:]name-version-release.arch

   int  beforeArch = envra.lastIndexOf('.');
   if (beforeArch <= 0) return false;

   int  beforeRel = envra.lastIndexOf('-', beforeArch-1);
   if (beforeRel <= 0) return false;

   int  beforeVer = envra.lastIndexOf('-', beforeRel-1);
   if (beforeVer <= 0) return false;

   int  beforeName = envra.lastIndexOf(':', beforeVer-1);
   // beforeName can be -1 legimitately

   if (beforeName == -1) {
      epoch = QString();
      // beforeName = -1 // coincidentally, it already is -1!
   } else {
      epoch = envra.mid(0, beforeName);
   }
   name    = envra.mid(beforeName+1, beforeVer - beforeName - 1);
   version = envra.mid(beforeVer+1, beforeRel - beforeVer - 1);
   release = envra.mid(beforeRel+1, beforeArch - beforeRel - 1);
   arch    = envra.mid(beforeArch+1);

#ifdef DEBUG_ENVRA
   printf("envra %s -> (n=%s, e=%s, v=%s, r=%s, a=%s)\n",
          envra.toStdString().c_str(),
          name.toStdString().c_str(),
          epoch.toStdString().c_str(),
          version.toStdString().c_str(),
          release.toStdString().c_str(),
          arch.toStdString().c_str());
#endif

   return true;
}

bool  YumIO::parseEnvra (QString   envra,
                         QString&  name,
                         QString&  arch)
{
   QString  epoch;
   QString  version;
   QString  release;
   return parseEnvra(envra, name, epoch, version, release, arch);
}

bool  YumIO::applyGroups()
{
   Package::EntryList::const_iterator  it = list.begin();
   for (; it != list.end(); ++it) {
      Package::Entry*  entry = *it;

      if (entry->kind == Package::GROUP) {
         switch (entry->apply.action) {
         case Package::UPDATE:
            if (!applyGrpUpdate(entry)) return false;
            break;
         case Package::INSTALL:
            if (!applyGrpInstall(entry)) return false;
            break;
         case Package::REMOVE:
            if (!applyGrpRemove(entry)) return false;
            break;
         case Package::NONE:
            // DO NOTHING
            break;
         }
      }
   }

   return true;
}

bool  YumIO::applyGrpUpdate (const Package::Entry*  grp)
{
   QSet<QString>  reqs;
   {
      PyObject*  reqPkgs = requiredPkgs(grp);
      if (reqPkgs == NULL) return false;
      PyObject*  it      = PyObject_GetIter(reqPkgs);
      for (;;) {
         PyObject*  pkgItem = PyIter_Next(it);
         if (pkgItem == NULL) break;
         reqs.insert(pyString(pkgItem));
         Py_DECREF(pkgItem);
      }
      Py_DECREF(it);
      Py_DECREF(reqPkgs);
   }
   
   PyObject*  pkgs = pkgTree(grp);
   if (pkgs == NULL) return false;
   PyObject*  it   = PyObject_GetIter(pkgs);
   for (;;) {
      PyObject*  pkgItem = PyIter_Next(it);
      if (pkgItem == NULL) break;

      Package::Entry*  entry = findPackage(pyString(pkgItem));
      if (entry != NULL) {
         switch (entry->apply.action) {
         case Package::NONE:
            // Only install required packages.  But if an optional package is
            // installed, do update it if necessary.
            switch (entry->state) {
            case Package::OUT_OF_DATE:
               entry->apply.action = Package::UPDATE;
               applyGrpLog(grp, entry);
               break;
            case Package::NOT_INSTALLED:
               if (reqs.contains(entry->name)) {
                  entry->apply.action = Package::INSTALL;
                  applyGrpLog(grp, entry);
               } else {
                  applyGrpLogNot(grp, entry, Package::INSTALL,
                                 tr("the package is optional"));
               }
               break;
            case Package::INSTALLED:
               // DO NOTHING
            default:
               // DO NOTHING
               break;
            }
            break;
         case Package::UPDATE:
         case Package::INSTALL:
            // DO NOTHING: no conflict
            applyGrpLogWas(grp, entry);
            break;
         case Package::REMOVE:
         default:
            if (!applyGrpConflict(grp, entry)) {
               return false;
            }
            break;
         }
      }

      Py_DECREF(pkgItem);
   }
   Py_DECREF(it);
   Py_DECREF(pkgs);

   return true;
}

bool  YumIO::applyGrpInstall (const Package::Entry*  grp)
{
   PyObject*  pkgs = pkgTree(grp);
   if (pkgs == NULL) return false;
   PyObject*  it   = PyObject_GetIter(pkgs);
   for (;;) {
      PyObject*  pkgItem = PyIter_Next(it);
      if (pkgItem == NULL) break;

      
      Package::Entry*  entry = findPackage(pyString(pkgItem));
      if (entry != NULL) {
         switch (entry->apply.action) {
         case Package::NONE:
            // Install or update all packages
            switch (entry->state) {
            case Package::OUT_OF_DATE:
               entry->apply.action = Package::UPDATE;
               applyGrpLog(grp, entry);
               break;
            case Package::NOT_INSTALLED:
               entry->apply.action = Package::INSTALL;
               applyGrpLog(grp, entry);
               break;
            case Package::INSTALLED:
            default:
               // DO NOTHING
               break;
            }
            break;
         case Package::UPDATE:
         case Package::INSTALL:
            // DO NOTHING: no conflicts
            applyGrpLogWas(grp, entry);
            break;
         case Package::REMOVE:
         default:
            if (!applyGrpConflict(grp, entry)) {
               return false;
            }
            break;
         }
      }

      Py_DECREF(pkgItem);
   }
   Py_DECREF(it);
   Py_DECREF(pkgs);

   return true;
}

bool  YumIO::applyGrpRemove (const Package::Entry*  grp)
{
   PyObject*  pkgs = allPkgs(grp);
   PyObject*  it   = PyObject_GetIter(pkgs);
   for (;;) {
      PyObject*  pkgItem = PyIter_Next(it);
      if (pkgItem == NULL) break;

      Package::Entry*  entry = findPackage(pyString(pkgItem));
      if (entry != NULL) {
         switch (entry->apply.action) {
         case Package::NONE:
            switch (entry->state) {
            case Package::OUT_OF_DATE:
            case Package::INSTALLED:
               entry->apply.action = Package::REMOVE;
               applyGrpLog(grp, entry);
               break;
            case Package::NOT_INSTALLED:
            default:
               // DO NOTHING
               break;
            }
            break;
         case Package::REMOVE:
            // DO NOTHING: no conflict
            applyGrpLog(grp, entry);
            break;
         case Package::UPDATE:
         case Package::INSTALL:
         default:
            if (!applyGrpConflict(grp, entry)) {
               return false;
            }
            break;
         }
      }

      Py_DECREF(pkgItem);
   }
   Py_DECREF(it);
   Py_DECREF(pkgs);

   return true;
}

bool  YumIO::applyGrpConflict (const Package::Entry*  grp,
                               Package::Entry*        pkg)
{
   Package::Action  grpPkgAction;

   switch (pkg->state) {
   case Package::OUT_OF_DATE:   grpPkgAction = Package::UPDATE;  break;
   case Package::NOT_INSTALLED: grpPkgAction = Package::INSTALL; break;
   case Package::INSTALLED:     grpPkgAction = Package::REMOVE;  break;
   default:
      error("Internal error: applyGrpConflict unexpected package state\n");
      return false;
   }

   QString  grpImage    = ACTION_IMAGE(grp->apply.action);
   QString  grpPkgImage = ACTION_IMAGE(grpPkgAction);
   QString  pkgImage    = ACTION_IMAGE(pkg->apply.action);

   QString  msg;
   if (pkg->apply.reason == pkg) {
      msg = tr("Package %1 is included in group %2 which was marked for %3,\n"
               "so the package would be marked for %4,\n"
               "but it was marked individually for %5.\n"
               "How do you wish to proceed?")
               .arg(pkg->name)
               .arg(grp->name)
               .arg(grpImage)
               .arg(grpPkgImage)
               .arg(pkgImage);
   } else if (pkg->apply.reason != NULL) {
      msg = tr("Package %1 is included in group %2 which was marked for %3,\n"
               "so the package would be marked for %4,\n"
               "but it also was part of group %5 and so was marked for %6.\n"
               "How do you wish to proceed?")
               .arg(pkg->name)
               .arg(grp->name)
               .arg(grpImage)
               .arg(grpPkgImage)
               .arg(pkg->apply.reason->name)
               .arg(pkgImage);
   } else {
      msg = tr("Package %1 is included in group %2 which was marked for %3,\n"
               "so the package would be marked for %4,\n"
               "but you resolved an earlier conflict by marking it for %5.\n"
               "How do you wish to proceed?")
               .arg(pkg->name)
               .arg(grp->name)
               .arg(grpImage)
               .arg(grpPkgImage)
               .arg(pkgImage);
   }

   int  result = dialogs->critical(/* caption             => */ tr("Apply"),
                                   /* text                => */ msg,
                                   /* button0Text         => */ grpPkgImage,
                                   /* button1Text         => */ pkgImage,
                                   /* button2Text         => */ tr("Cancel"),
                                   /* defaultButtonNumber => */ 2,
                                   /* escapeButtonNumber  => */ 2);
   
   switch (result) {
   case 0:
      pkg->apply.action = grpPkgAction;
      break;
   case 1:
      // DO NOTHING
      break;
   case 2:
      return false;
   }

   log(Log::NORMAL, 4, tr("Manually resolving conflict: marking package %1 for"
                          " %3").arg(pkg->name)
                                .arg(ACTION_IMAGE(pkg->apply.action)));
   
      
   pkg->apply.reason = NULL; // to indicate a user conflict resolution
   return true;
}

void  YumIO::applyGrpLog (const Package::Entry*  grp,
                          const Package::Entry*  pkg)
{
   log(Log::NORMAL, 4, tr("For group %1, marking package %2 for"
                          " %3").arg(grp->name)
                                .arg(pkg->name)
                                .arg(ACTION_IMAGE(pkg->apply.action)));
}

void  YumIO::applyGrpLogWas (const Package::Entry*  grp,
                             const Package::Entry*  pkg)
{
   log(Log::NORMAL, 4, tr("For group %1, package %2 already was marked for"
                          " %3").arg(grp->name)
                                .arg(pkg->name)
                                .arg(ACTION_IMAGE(pkg->apply.action)));
}

void  YumIO::applyGrpLogNot (const Package::Entry*  grp,
                             const Package::Entry*  pkg,
                             Package::Action        action,
                             QString                reason)
{
   log(Log::NORMAL, 4, tr("For group %1, not marking package %2 for"
                          " %3, becuase %4").arg(grp->name)
                                            .arg(pkg->name)
                                            .arg(ACTION_IMAGE(action))
                                            .arg(reason));
}

PyObject*  YumIO::requiredPkgs (const Package::Entry*  grp)
{
   if (hasDistribGroups()) {
      PyObject*  mand = PyObject_GetAttrString(grp->obj, "mandatory_packages");
      PyObject*  def  = PyObject_GetAttrString(grp->obj, "default_packages");
      if (mand == NULL || def == NULL) {
         error(tr("Internal error: group lacks required package lists"));
         return NULL;
      }
      PyObject*  mand_keys = pyMethodCall(mand, "keys", "()");
      PyObject*  def_keys  = pyMethodCall(def,  "keys", "()");
      if (mand_keys == NULL || def_keys == NULL) {
         error(tr("Internal error:"
                  " group lacks required proper required package lists"));
         return NULL;
      }
      PyObject*  result = PySequence_Concat(mand_keys, def_keys);
      Py_DECREF(def_keys);
      Py_DECREF(mand_keys);
      Py_DECREF(def);
      Py_DECREF(mand);
      return result;
   } else {
      char*      name   = strdup(grp->name.toStdString().c_str());
      PyObject*  result = pyMethodCall(grp->obj, "requiredPkgs", "(s)", name);
      free(name);
      return result;
   }
}

PyObject*  YumIO::pkgTree (const Package::Entry*  grp)
{
   if (hasDistribGroups()) {
      // There are no subgroups in yum >= 2.5, so these two are equivalent.
      return requiredPkgs(grp);
   } else {
      char*      name   = strdup(grp->name.toStdString().c_str());
      PyObject*  result = pyMethodCall(grp->obj, "pkgTree", "(s)", name);
      free(name);
      return result;
   }
}

PyObject*  YumIO::allPkgs (const Package::Entry*  grp)
{
   if (hasDistribGroups()) {
      PyObject*  mand = PyObject_GetAttrString(grp->obj, "mandatory_packages");
      PyObject*  def  = PyObject_GetAttrString(grp->obj, "default_packages");
      PyObject*  opt  = PyObject_GetAttrString(grp->obj, "optional_packages");
      PyObject*  cond = PyObject_GetAttrString(grp->obj, 
                                               "conditional_packages");
      if (mand == NULL || def == NULL || opt == NULL || cond == NULL) {
         error(tr("Internal error: group lacks package lists"));
         return NULL;
      }
      PyObject*  mand_keys = pyMethodCall(mand, "keys", "()");
      PyObject*  def_keys  = pyMethodCall(def,  "keys", "()");
      PyObject*  opt_keys  = pyMethodCall(opt,  "keys", "()");
      PyObject*  cond_keys = pyMethodCall(cond, "keys", "()");
      if (mand_keys == NULL || def_keys == NULL
          || opt_keys == NULL || cond_keys == NULL) {
         error(tr("Internal error:"
                  " group lacks required proper package lists"));
         return NULL;
      }
      PyObject*  temp1  = PySequence_Concat(mand_keys, def_keys);
      PyObject*  temp2  = PySequence_Concat(opt_keys,  cond_keys);
      PyObject*  result = PySequence_Concat(temp1, temp2);
      Py_DECREF(temp2);
      Py_DECREF(temp1);
      Py_DECREF(cond_keys);
      Py_DECREF(opt_keys);
      Py_DECREF(def_keys);
      Py_DECREF(mand_keys);
      Py_DECREF(cond);
      Py_DECREF(opt);
      Py_DECREF(def);
      Py_DECREF(mand);
      return result;
   } else {
      char*      name   = strdup(grp->name.toStdString().c_str());
      PyObject*  result = pyMethodCall(grp->obj, "requiredPkgs", "(s)", name);
      free(name);
      return result;
   }
}

bool  YumIO::makeLists (PyObject*   yumIO,
                        PyObject*&  tsInfo,
                        PyObject*&  lists,
                        PyObject*&  updated,
                        PyObject*&  installed,
                        PyObject*&  removed,
                        PyObject*&  obsoleted,
                        PyObject*&  depup,
                        PyObject*&  depinst,
                        PyObject*&  deprem)
{
   tsInfo = PyObject_GetAttrString(yumIO, "tsInfo");
   lists  = pyMethodCall(tsInfo, "makelists", "()");

   if (yumVersion() >= 20302) {
      updated   = PyObject_GetAttrString(tsInfo, "updated");
      installed = PyObject_GetAttrString(tsInfo, "installed");
      removed   = PyObject_GetAttrString(tsInfo, "removed");
      obsoleted = PyObject_GetAttrString(tsInfo, "obsoleted");
      depup     = PyObject_GetAttrString(tsInfo, "depupdated");
      depinst   = PyObject_GetAttrString(tsInfo, "depinstalled");
      deprem    = PyObject_GetAttrString(tsInfo, "depremoved");
      // Make these objects all be borrowed refs to conform with the older
      // yum case, so that the caller only needs to deal with their
      // reference counts in one way.
      Py_DECREF(updated);
      Py_DECREF(installed);
      Py_DECREF(removed);
      Py_DECREF(obsoleted);
      Py_DECREF(depup);
      Py_DECREF(depinst);
      Py_DECREF(deprem);
   } else {
      if (!PyTuple_Check(lists)) {
         error("Internal error: yumIO.makelists() did not return a tuple\n");
         Py_DECREF(tsInfo);
         Py_DECREF(lists);
         return false;
      }
      if (PyTuple_Size(lists) != 7) {
         error("Internal error: yumIO.makelists() returned tuple of incorrect"
               " size\n");
         Py_DECREF(tsInfo);
         Py_DECREF(lists);
         return false;
      }
      updated   = PyTuple_GetItem(lists, 0);
      installed = PyTuple_GetItem(lists, 1);
      removed   = PyTuple_GetItem(lists, 2);
      obsoleted = PyTuple_GetItem(lists, 3);
      depup     = PyTuple_GetItem(lists, 4);
      depinst   = PyTuple_GetItem(lists, 5);
      deprem    = PyTuple_GetItem(lists, 6);
   }

   return true;
}

QString  YumIO::simpleString (PyObject*    pkg,
                              const char*  key)
{
   char*  copy = new char[strlen(key)+1];
   strcpy(copy, key);
   QString  result = simpleString(pkg, copy);
   delete [] copy;
   return result;
}

QString  YumIO::simpleString (PyObject*  pkg,
                              char*      key)
{
   Python::PyGILStateWrapper  gsw;

   pyExceptStart();
   pyExceptExpect(PyExc_KeyError);
   PyObject*  obj = pyMethodCall(pkg, "returnSimple", "(s)", key);
   pyExceptClear();
   if (obj == NULL || obj == Py_None) return QString();

   QString  result = pyStringXDec(obj);
   result.remove(QRegExp("\n*$"));

   return result;
}

QString  YumIO::getBiArch(QString  arch)
{
   QString  result;

   char*  arch_s = strdup(arch.toStdString().c_str());
   result = pyStringDec(pyMethodCall(yumIO, "getBiArch", "(s)", arch_s));
   free(arch_s);

   return result;
}

Package::Size  YumIO::sizeOf(PyObject*  pkg)
{
   PyObject*      sizeObj = PyObject_GetAttrString(pkg, "size");
   Package::Size  result  = 0;

   if (PyLong_Check(sizeObj)) {
      result = PyLong_AsLongLong(sizeObj);
   } else if (PyInt_Check(sizeObj)) {
      result = PyInt_AsLong(sizeObj);
   } else if (PyString_Check(sizeObj)) {
      QString  str = pyString(sizeObj);
      errno = 0;
      result = strtoull(str.toStdString().c_str(), NULL, 10);
      if (errno != 0) {
         warning(tr("Internal error: package size not understood:"
                    " %1").arg(str));
      }
   } else if (PyCallable_Check(sizeObj)) {
      PyObject*  obj = pyMethodCall(pkg, "size", "()");

      if (PyLong_Check(obj)) {
         result = PyLong_AsLongLong(obj);
         Py_DECREF(obj);
      } else if (PyInt_Check(obj)) {
         result = PyInt_AsLong(obj);
         Py_DECREF(obj);
      } else if (PyString_Check(obj)) {
         QString  str = pyStringDec(obj);
         errno = 0;
         result = strtoull(str.toStdString().c_str(), NULL, 10);
         if (errno != 0) {
            warning(tr("Internal error: package size not understood:"
                       " %1").arg(str));
         }
      } else {
         warning(tr("Internal error: package size result not understood"));
         Py_XDECREF(obj);
      }
   } else {
      warning(tr("Internal error: package size not understood"));
   }
   
   Py_DECREF(sizeObj);

   return result;
}

QString  YumIO::bytesFormat(Package::Size  bytes)
{
   QString  result;
   double   units  = (double)bytes;
   QString  suffix = tr("b");

   if (bytes >= 1000ULL*1024*1024*1024*1024*1024) {
      units /= 1024.0*1024*1024*1024*1024*1024;
      suffix = tr("Eb");
   } else if (bytes >= 1000ULL*1024*1024*1024*1024) {
      units /= 1024.0*1024*1024*1024*1024;
      suffix = tr("Pb");
   } else if (bytes >= 1000ULL*1024*1024*1024) {
      units /= 1024.0*1024*1024*1024;
      suffix = tr("Tb");
   } else if (bytes >= 1000ULL*1024*1024) {
      units /= 1024.0*1024*1024;
      suffix = tr("Gb");
   } else if (bytes >= 1000ULL*1024) {
      units /= 1024.0*1024;
      suffix = tr("Mb");
   } else if (bytes >= 1000ULL) {
      units /= 1024.0;
      suffix = tr("Kb");
   }

   if (0.0 < units && units < 9.95) {
      result.sprintf("%.1f", units);
   } else {
      result.sprintf("%.0f", units);
   }
   result.append(suffix);

   return result;
}

time_t  YumIO::buildTimeOf(PyObject*  pkg)
{
   return pyTimeDec(pyMethodCall(pkg, "returnSimple", "(s)", "buildtime"));
}

#define TIME_1  "Mon Dec 31 "
#define TIME_2  "Mon Dec 31 03:00:00 "

char*  YumIO::dateFormat(time_t  secs)
{
   static char  buffer[1024];
   ctime_r(&secs, buffer);
   buffer[strlen(buffer)-1] = '\0'; // Strip the newline
   strcpy(&buffer[STRLEN(TIME_1)], &buffer[STRLEN(TIME_2)]);
   return buffer;
}

char*  YumIO::timeFormat(time_t  secs)
{
   static char  buffer[1024];
   ctime_r(&secs, buffer);
   buffer[strlen(buffer)-1] = '\0'; // Strip the newline
   return buffer;
}

QString  YumIO::yumVersionStr()
{
   if (!yumVerKnown) {
      PyObject*  ver = pyValue("yum", "__version__");
      if (ver == NULL) {
         ver = pyValue("cli", "__version__");
      }

      if (ver != NULL) {
         yumVer = pyString(ver);
      }
      yumVerKnown = true;
   }

   return yumVer;
}

long  YumIO::yumVersion()
{
   QStringList  versions = yumVersionStr().split('.');
   long         result   = 0;

   int   i;
   long  mult;
   for (i = 0, mult = 10000; i < 3; i++, mult /= 100) {
      if (i >= versions.count()) break;

      errno = 0;
      long  value = strtoul(versions[i].toStdString().c_str(), NULL, 10);
      if (errno != 0) {
         static bool seen = false;
         if (!seen) {
            warning(tr("Internal error: yum version not understood: %1"
                       " errno=%2 (%3)")
                    .arg(yumVersionStr()).arg(errno).arg(strerror(errno)));
            seen = true;
         }
      }

      result += value * mult;
   }

   return result;
}

bool  YumIO::hasDistribGroups()
{
   // This implies that group information now is distributed into a separate
   // class for each individual group.  It also implies that the concept of
   // sub-groups (and metapkgs) have been removed.
   return yumVersion() >= 20500;
}

bool  YumIO::hasTransTest()
{
   return yumVersion() < 30203;
}

bool  YumIO::usesSqlite()
{
   if (!usesSqliteKnown) {
      pyExceptStart();
      usesSqliteValue = (pyCall("usesSqlite", "()") == Py_True);
      pyExceptClear();
      usesSqliteKnown = true;
   }
   return usesSqliteValue;
}

bool  YumIO::usesPresto()
{
   if (!usesPrestoKnown) {
      pyExceptStart();
      usesPrestoValue = (pyMethodCall(yumIO, "usesPresto", "()") == Py_True);
      pyExceptClear();
      usesPrestoKnown = true;
   }
   return usesPrestoValue;
}

bool  YumIO::hasMultiProgress()
{
   if (!hasMultiProgressKnown) {
      pyExceptStart();
      hasMultiProgressValue = (pyMethodCall(yumIO, "hasMultiProgress", "()")
                               == Py_True);
      pyExceptClear();
      hasMultiProgressKnown = true;
   }
   return hasMultiProgressValue;
}

bool  YumIO::hasPickledCache()
{
   return yumVersion() < 20902;
}

bool  YumIO::hasSqliteCache()
{
   return yumVersion() >= 20302;
}

QString  YumIO::pyString(PyObject*  obj)
{
   if (PyString_Check(obj)) {
      return Python::pyString(obj);
   } else if (PyUnicode_Check(obj)) {
      Py_ssize_t  size   = PyUnicode_GetSize(obj);
      wchar_t*    buffer = new wchar_t[size+1];
      Py_ssize_t  actual = PyUnicode_AsWideChar((PyUnicodeObject*)obj,
                                                buffer, size+1);
      if (actual == -1 && PyErr_Occurred()) handleException();
      if (actual >= 0) {
         buffer[actual] = '\0';
      }
      QString  result = QString::fromStdWString(buffer);
      delete [] buffer;
      return result;
   } else if (obj == Py_None) {
      return QString();
   } else {
      warning(tr("Internal error: YumIO::pyString did not understand type"));
      return QString();
   }
}

QString  YumIO::pyStringDec(PyObject*  obj)
{
   QString  result = pyString(obj);
   Py_XDECREF(obj);
   return result;
}

// Generally the same as pyString.  But if a yum string is parenthesized
// (e.g. "(null)"), then it gets turned into a list.  This copes with that and
// turns it back into a string.  And to support pythonExceptionV, it also
// translates a variety of lists, tuples, and integers.
QString  YumIO::pyStringX (PyObject*  obj)
{
   if (PyString_Check(obj) || PyUnicode_Check(obj)) {
      return pyString(obj);
   } else if (PyList_Check(obj)) {
      if (PyList_Size(obj) == 0) {
         return QString();
      } else {
         QString  result("(");
         PyObject*  it = PyObject_GetIter(obj);
         bool       first = true;
         for (;;) {
            PyObject*  item = PyIter_Next(it);
            if (item == NULL) break;
            if (!first) result.append(",\n ");
            result.append(pyStringX(item));
            Py_DECREF(item);
            first = false;
         }
         Py_DECREF(it);
         result.append(")");
         return result;
      }
   } else if (PyTuple_Check(obj)) {
      if (PyTuple_Size(obj) == 0) {
         return QString();
      } else {
         QString  result("(");
         PyObject*  it = PyObject_GetIter(obj);
         bool       first = true;
         for (;;) {
            PyObject*  item = PyIter_Next(it);
            if (item == NULL) break;
            if (!first) result.append(',');
            result.append(pyStringX(item));
            Py_DECREF(item);
            first = false;
         }
         Py_DECREF(it);
         result.append(")");
         return result;
      }
   } else if (PyLong_Check(obj)) {
      return QString::number(PyLong_AsLongLong(obj));
   } else if (PyInt_Check(obj)) {
      return QString::number(PyInt_AsLong(obj));
   } else {
      warning(tr("Internal error: YumIO::pyStringX did not understand type"));
      return QString();
   }      
}

QString  YumIO::pyStringXDec (PyObject*  obj)
{
   QString  result = pyStringX(obj);
   Py_XDECREF(obj);
   return result;
}

QString  YumIO::pyMappingString (PyObject*    obj,
                                 const char*  attr)
{
   char*  copy = new char[strlen(attr)+1];
   strcpy(copy, attr);
   QString  result = pyStringDec(PyMapping_GetItemString(obj, copy));
   delete [] copy;
   return result;
}

time_t  YumIO::pyTime(PyObject*  obj)
{
   time_t  result = 0;

   if (PyInt_Check(obj)) {
      result = PyInt_AsLong(obj);
   } else if (PyLong_Check(obj)) {
      result = PyLong_AsLongLong(obj);
   } else if (PyString_Check(obj)) {
      QString  str = pyString(obj);
      errno = 0;
      result = strtoul(str.toStdString().c_str(), NULL, 10);
      if (errno != 0) {
         warning(tr("Internal error: time not understood: %1").arg(str));
      }
   } else {
      warning(tr("Internal error: time not understood"));
   }

   return result;
}

time_t  YumIO::pyTimeDec(PyObject*  obj)
{
   time_t  result = 0;

   if (PyInt_Check(obj)) {
      result = PyInt_AsLong(obj);
      Py_DECREF(obj);
   } else if (PyLong_Check(obj)) {
      result = PyLong_AsLongLong(obj);
      Py_DECREF(obj);
   } else if (PyString_Check(obj)) {
      QString  str = pyStringDec(obj);
      errno = 0;
      result = strtoul(str.toStdString().c_str(), NULL, 10);
      if (errno != 0) {
         warning(tr("Internal error: time not understood: %1").arg(str));
      }
   } else {
      warning(tr("Internal error: time not understood"));
      Py_DECREF(obj);
   }

   return result;
}

PyObject*  YumIO::pyDictGetItem (PyObject*       dict,
                                 const QString&  key)
{
   // This is because PyDict_GetItemString's 2nd parameter is a char* and not
   // a const char*.  It probably is just that its parameter profile isn't
   // const-correct.  But I'm not chancing it.
   char*      keyCopy = strdup(key.toStdString().c_str());
   PyObject*  result  = PyDict_GetItemString(dict, keyCopy);
   free(keyCopy);
   return result;
}

const char*  YumIO::pyDescribe (PyObject*  obj)
{
   if (PyString_Check(obj) || PyUnicode_Check(obj)) {
      return pyString(obj).toStdString().c_str();
   } else if (PyLong_Check(obj)) {
      static char  buffer[1024];
      sprintf(buffer, "%lld", PyLong_AsLongLong(obj));
      return buffer;
   } else if (PyInt_Check(obj)) {
      static char  buffer[1024];
      sprintf(buffer, "%ld", PyInt_AsLong(obj));
      return buffer;
   } else if (obj == Py_None) {
      return "*None*";
   } else {
      PyObject*    str    = pyMethodCall(obj, "__str__", "()");
      const char*  result = "*Obj-no-str*";
      if (PyString_Check(str) || PyUnicode_Check(str)) {
         static char  buffer[1024];
         sprintf(buffer, "%s(*Obj*)", pyString(str).toStdString().c_str());
         result = buffer;
      }
      Py_XDECREF(str);
      return result;
   }
}

void  YumIO::log (Log::Kind     kind,
                  unsigned int  value,
                  QString       text)
{
   QString  edited(text);
   edited.replace(QRegExp(": (\\[Errno [0-9]*\\] )"), ":\n\\1");

   QStringList  lines = edited.split("\n");
   
   logEntries.lock();

   QStringList::const_iterator  it = lines.begin();
   for (; it != lines.end(); ++it) {
      logEntries.append(Log::Entry(kind, value, *it));
   }
   
   logEntries.unlock();
   logEntries.emitAppended();
}

bool  YumIO::reapChildren ()
{
   // If a previous apply was aborted somehow, there may still be applydeltarpm
   // children in zombie state.  Reap them so they are not discovered by the
   // next apply in yum.drpm.DeltaInfo._wait, where the exit of an unexpected
   // child can cause an assertion failure.

   for (;;) {
      pid_t  child = waitpid(-1, NULL, WNOHANG);
      if (child == 0)                     break;
      if (child == -1 && errno == ECHILD) break;
      if (child == -1) {
         error(tr("Unable to reap children; errno=%1 (%2)")
                  .arg(errno).arg(strerror(errno)));
         return false;
      }
      log(Log::WARNING, 0, tr("Reaped zombie child %1").arg(child));
   }

   return true;
}

void  YumIO::sigchldDisable ()
{
   struct sigaction  noAct;
   noAct.sa_handler = SIG_DFL;
   sigemptyset(&noAct.sa_mask);
   noAct.sa_flags   = 0;
   int  status = sigaction(SIGCHLD, &noAct, &savedSigChldAct);
   if (status == -1) {
      error(tr("Unable to disable SIGCHLD signal handler; errno=%1 (%2)")
               .arg(errno).arg(strerror(errno)));
   }
}

void  YumIO::sigchldRestore ()
{
   int  status = sigaction(SIGCHLD, &savedSigChldAct, NULL);
   if (status == -1) {
      error(tr("Unable to restore SIGCHLD signal handler; errno=%1 (%2)")
               .arg(errno)
               .arg(strerror(errno)));
   }
}

void  YumIO::commExceptSetup ()
{
   pyExceptExpect(xcpt.bdbQuit);
   pyExceptExpect(xcpt.yumBaseError);
   pyExceptExpect(xcpt.repoError);
   pyExceptExpect(PyExc_IOError);
   pyExceptExpect(PyExc_KeyboardInterrupt);
   pyExceptExpect(PyExc_AssertionError);

   if (xcpt.sqlite3OperationalError != NULL) {
      pyExceptExpect(xcpt.sqlite3OperationalError);
   }
   if (xcpt.noMoreMirrorsRepoError != NULL) {
      pyExceptExpect(xcpt.noMoreMirrorsRepoError);
   }
}

bool  YumIO::commExceptRaised (PyObject*  result,
                               QString    descr)
{
   // Be careful here.  You must not do pyLookup() in this routine, as that can
   // call PyErr_Clear(), which makes the pyExceptMatches() queries all return
   // false.

   if (pyExceptMatches(result, PyExc_IOError)) {
      PyObject*  exception;
      PyObject*  v;
      PyObject*  tb;
      PyErr_Fetch(&exception, &v, &tb);
      PyErr_NormalizeException(&exception, &v, &tb);
      PyErr_Display(exception, v, tb);
      QString  value = pyStringDec(pyExceptValuef());
      error(tr("Unable to %1: I/O Error: %2\n\n"
               "Emitting a walkback to the Log window.")
               .arg(descr).arg(value));
      return true;
   } else if (pyExceptMatches(result, PyExc_KeyboardInterrupt)) {
      PyObject*  exception;
      PyObject*  v;
      PyObject*  tb;
      PyErr_Fetch(&exception, &v, &tb);
      PyErr_NormalizeException(&exception, &v, &tb);
      PyErr_Display(exception, v, tb);
      error(tr("Unable to %1: KeyboardInterrupt\n\n"
               "This usually means NUU was sent signal SIGINT,"
               " SIGTERM, or similar.\n\n"
               "But urlgrabber also can raise this error for several"
               " situations where it is just guessing.\n"
               "So, emitting a walkback to the Log window.")
               .arg(descr));
      return true;
   } else if (pyExceptMatches(result, PyExc_AssertionError)) {
      PyObject*  exception;
      PyObject*  v;
      PyObject*  tb;
      PyErr_Fetch(&exception, &v, &tb);
      PyErr_NormalizeException(&exception, &v, &tb);
      PyErr_Display(exception, v, tb);
      error(tr("Internal error: python assertion failed\n"
               "When reporting this issue, please include the contents of the\n"
               "Log window, which should contain a walkback."));
      return true;
   } else if (xcpt.noMoreMirrorsRepoError != NULL
              && pyExceptMatches(result, xcpt.noMoreMirrorsRepoError)) {
      QString  value = pyStringDec(pyExceptValuef());
      error(tr("Unable to %1: metadata unavailable from all repo mirrors:\n"
               "%2").arg(descr).arg(value));
      return true;
   } else if (pyExceptMatches(result, xcpt.repoError)
              || (xcpt.sqlite3OperationalError != NULL
                  && pyExceptMatches(result, xcpt.sqlite3OperationalError))) {
      QString  value = pyStringDec(pyExceptValuef());
      if (value.contains("database is locked")) {
         pythonException(tr("Unable to %1: a repository's sqlite database"
                            " was locked unexpectedly\n"
                            "On some distributions, PackageKit has been known"
                            " to cause this.").arg(descr));
         return true;
      } else {
         exceptionError();
         return true;
      }
   } else if (pyExceptMatches(result, xcpt.bdbQuit)
              || pyExceptMatches(result, xcpt.yumBaseError)) {
      pythonExceptionV(tr("Unable to %1").arg(descr));
      return true;
   }

   return false;
}

void  YumIO::pyExceptExpect (PyObject*  xcpt)
{
   logEntries.lock();
   logStartOp = logEntries.count();
   logEntries.unlock();
   Python::pyExceptExpect(xcpt);
}

void  YumIO::pyExceptClear ()
{
   logStartOp = -1;
   Python::pyExceptClear();
}

void  YumIO::pythonException (QString  text)
{
   int  startOp;
   if (logStartOp != -1) {
      startOp = logStartOp;
   } else {
      logEntries.lock();
      startOp = logEntries.count();
      logEntries.unlock();
   }

   log(Log::ERROR, 0, text);

   QString  msg;

   logEntries.lock();
   int  i;
   for (i = startOp; i < logEntries.count(); i++) {
      if (logEntries[i].kind == Log::ERROR && logEntries[i].value <= 2) {
         if (!msg.isEmpty()) {
            msg.append('\n');
         }
         msg.append(logEntries[i].text);
      }
   }
   logEntries.unlock();

   dialogs->scrollable(tr("NUU Yum Error"), msg);
}

void  YumIO::pythonExceptionV (QString  text)
{
   int  startOp;
   if (logStartOp != -1) {
      startOp = logStartOp;
   } else {
      logEntries.lock();
      startOp = logEntries.count();
      logEntries.unlock();
   }

   log(Log::ERROR, 0, pyStringXDec(pyExceptValuef()));
   log(Log::ERROR, 0, text);

   QString  msg;

   logEntries.lock();
   int  i;
   for (i = startOp; i < logEntries.count(); i++) {
      if (logEntries[i].kind == Log::ERROR && logEntries[i].value <= 2) {
         if (!msg.isEmpty()) {
            msg.append('\n');
         }
         msg.append(logEntries[i].text);
      }
   }
   logEntries.unlock();

   dialogs->scrollable(tr("NUU Yum Error"), msg);
}

void  YumIO::exceptionExpected()
{
#ifdef DEBUG_EXPECTED_EXCEPTIONS
   // This exception was totally expected.  But maybe for debugging I'd
   // want to know the particulars.  So, stash this at a very high log
   // level to imply its debugging nature.  Note that a Search operation
   // will cause tons of these.

   Log::Kind     oldStderrKind  = wr.stderrKind;
   unsigned int  oldStderrValue = wr.stderrValue;

   wr.stderrKind  = Log::NORMAL;
   wr.stderrValue = 10;

   PyObject*  exception;
   PyObject*  v;
   PyObject*  tb;
   PyErr_Fetch(&exception, &v, &tb);
   PyErr_NormalizeException(&exception, &v, &tb);
   PyErr_Display(exception, v, tb);
   PyErr_Restore(exception, v, tb);

   wr.stderrKind  = oldStderrKind;
   wr.stderrValue = oldStderrValue;

   Python::exceptionExpected();
#endif
}

/*
** Python type interfaces
*/

QString  YumIO::pkgtupFormat(PyObject* pkgtup)
{
   QString  name  = pyString(PyTuple_GetItem(pkgtup, 0));
   QString  arch  = pyString(PyTuple_GetItem(pkgtup, 1));
   QString  epoch = pyString(PyTuple_GetItem(pkgtup, 2));
   QString  ver   = pyString(PyTuple_GetItem(pkgtup, 3));
   QString  rel   = pyString(PyTuple_GetItem(pkgtup, 4));

   if (epoch == QString("0")) {
      return tr("%1 %3-%4 (%2)").arg(name).arg(arch).arg(ver).arg(rel);
   } else {
      return tr("%1 %3:%4-%5 (%2)")
                .arg(name).arg(arch).arg(epoch).arg(ver).arg(rel);
   }
}

QString  YumIO::pkgtupRepoFormat(PyObject*  pkgtup,
                                 QString    repoid)
{
   return pkgtupFormat(pkgtup).append(tr(" [%1]").arg(repoid));
}

int YumIO::displayTransactionPkgs(QString    action,
                                  PyObject*  txmbrs)
{
   PyObject*  it = PyObject_GetIter(txmbrs);
   for (;;) {
      PyObject*  txmbr = PyIter_Next(it);
      if (txmbr == NULL) break;
      PyObject*  pkgtup = PyObject_GetAttrString(txmbr, "pkgtup");
      if (txmbr == NULL) {
         error("Internal error: txmbr lacks pkgtup\n");
         Py_DECREF(txmbr);
         Py_DECREF(it);
         return 0;
      }
      ts.descr.append(tr("\t%1: %2\n").arg(action).arg(pkgtupFormat(pkgtup)));
      Py_DECREF(pkgtup);
      Py_DECREF(txmbr);
   }

   Py_DECREF(it);

   return 1;
}

int YumIO::displayTransactionPkgsRepo(QString    action,
                                      PyObject*  txmbrs)
{
   PyObject*  it = PyObject_GetIter(txmbrs);
   for (;;) {
      PyObject*  txmbr = PyIter_Next(it);
      if (txmbr == NULL) break;
      PyObject*  pkgtup = PyObject_GetAttrString(txmbr, "pkgtup");
      if (txmbr == NULL) {
         error("Internal error: txmbr lacks pkgtup\n");
         Py_DECREF(txmbr);
         Py_DECREF(it);
         return 0;
      }
      QString  repoid = pyStringDec(PyObject_GetAttrString(txmbr, "repoid"));
      ts.descr.append(tr("\t%1: %2\n").arg(action)
                                      .arg(pkgtupRepoFormat(pkgtup, repoid)));
      Py_DECREF(pkgtup);
      Py_DECREF(txmbr);
   }

   Py_DECREF(it);

   return 1;
}

PyObject*  YumIO::findRelation(PyObject*    relatedto,
                               const char*  relation)
{
   PyObject*  it = PyObject_GetIter(relatedto);
   for (;;) {
      PyObject*  tuple = PyIter_Next(it);
      if (tuple == NULL) break;

      if (strcmp(Python::pyString(PyTuple_GetItem(tuple, 1)), relation) == 0) {
         PyObject*  result = PyTuple_GetItem(tuple, 0);
         // Make result a new reference
         Py_INCREF(result);
         Py_DECREF(tuple);
         Py_DECREF(it);
         return result;
      }

      Py_DECREF(tuple);
   }

   Py_DECREF(it);

   return NULL;
}

int YumIO::displayTransactionPkgsObs(QString    action,
                                     PyObject*  txmbrs)
{
   PyObject*  it = PyObject_GetIter(txmbrs);
   for (;;) {
      PyObject*  txmbr = PyIter_Next(it);
      if (txmbr == NULL) break;
      PyObject*  pkgtup = PyObject_GetAttrString(txmbr, "pkgtup");
      if (txmbr == NULL) {
         error("Internal error: txmbr lacks pkgtup\n");
         return 0;
      }
      PyObject*  relatedto = PyObject_GetAttrString(txmbr, "relatedto");
      if (relatedto == NULL) {
         error("Internal error: package has no relatedto\n");
         Py_DECREF(pkgtup);
         Py_DECREF(txmbr);
         Py_DECREF(it);
         return 0;
      }
      PyObject*  obspkg = findRelation(relatedto, "obsoletedby");
      if (obspkg == NULL) {
         error("Internal error: couldn't find obsoletedby package\n");
         Py_DECREF(relatedto);
         Py_DECREF(pkgtup);
         Py_DECREF(txmbr);
         Py_DECREF(it);
         return 0;
      }
      if (yumVersion() >= 20302) {
         PyObject*  obstup = PyObject_GetAttrString(obspkg, "pkgtup");
         if (obspkg == NULL) {
            error("Internal error: couldn't find obsoletedby package\n");
            Py_DECREF(obspkg);
            Py_DECREF(relatedto);
            Py_DECREF(pkgtup);
            Py_DECREF(txmbr);
            Py_DECREF(it);
            return 0;
         }
         ts.descr.append(tr("\t%1: %2 with %3\n").arg(action)
                                                 .arg(pkgtupFormat(pkgtup))
                                                 .arg(pkgtupFormat(obstup)));
         Py_DECREF(obstup);
      } else {
         ts.descr.append(tr("\t%1: %2 with %3\n").arg(action)
                                                 .arg(pkgtupFormat(pkgtup))
                                                 .arg(pkgtupFormat(obspkg)));
      }
      Py_DECREF(obspkg);
      Py_DECREF(relatedto);
      Py_DECREF(pkgtup);
      Py_DECREF(txmbr);
   }

   Py_DECREF(it);

   return 1;
}

void  YumIO::noteTransactionPkgs(Package::Action  action,
                                 PyObject*        txmbrs)
{
   PyObject*  it = PyObject_GetIter(txmbrs);

   for (;;) {
      PyObject*  txmbr = PyIter_Next(it);
      if (txmbr == NULL) break;
      PyObject*  pkgtup = PyObject_GetAttrString(txmbr, "pkgtup");
      if (txmbr == NULL) {
         error("Internal error: txmbr lacks pkgtup\n");
         Py_DECREF(txmbr);
         Py_DECREF(it);
         return;
      }

      QString          name  = pyString(PyTuple_GetItem(pkgtup, 0));
      QString          arch  = pyString(PyTuple_GetItem(pkgtup, 1));
      Package::Entry*  entry = hash[EntryKey(false, name, getBiArch(arch))];
      switch (action) {
      case Package::INSTALL:
         ts.run.nominalGrabSize += PG_R_GRAB(entry->size);
         if (!QFile::exists(getLocalPkg(entry->obj))) {
            ts.run.totalGrabSize += PG_R_GRAB(entry->size);
         }
         ts.run.totalTsSize += PG_R_INSTALL(entry->size);
         if (entry->apply.action == Package::NONE) {
            entry->apply.action = Package::INSTALL;
         }
         break;
      case Package::UPDATE:
         ts.run.nominalGrabSize += PG_R_GRAB(entry->size);
         if (!QFile::exists(getLocalPkg(entry->obj))) {
            ts.run.totalGrabSize += PG_R_GRAB(entry->size);
         }
         ts.run.totalTsSize += PG_R_UPDATE(entry->size);
         ts.run.totalTsSize += PG_R_COMPLETE_UPDATE(entry->next->size);
         if (entry->apply.action == Package::NONE) {
            entry->apply.action = Package::UPDATE;
         }
         break;
      case Package::REMOVE:
         ts.run.totalTsSize += PG_R_REMOVE(entry->size);
         if (entry->apply.action == Package::NONE) {
            entry->apply.action = Package::REMOVE;
         }
         break;
      default:
         // DO NOTHING
         break;
      }

      Py_DECREF(pkgtup);
      Py_DECREF(txmbr);
   }

   Py_DECREF(it);
}

void  YumIO::error(QString  text)
{
   log(Log::ERROR, 0, text);
   dialogs->critical(tr("NUU Yum Error"), text);
}

void  YumIO::warning(QString  text)
{
   log(Log::WARNING, 0, text);
   dialogs->warning(tr("NUU Yum Warning"), text);
}

void  YumIO::error(const char*  text)
{
   error(QString(text));
}


void  YumIO::warning(const char*  text)
{
   warning(QString(text));
}
